r"""
CL4KT
################################################
Reference:

    Wongsung Lee et al. "Contrastive Learning for Knowledge Tracing." in WWW2022.

Reference Code:

    https://github.com/UpstageAI/cl4kt
"""
from ..gd_basemodel import GDBaseModel
import random
import math
import torch.nn as nn
import torch
import torch.nn.functional as F
from edustudio.utils.common import set_same_seeds
import numpy as np
from ...datatpl.utils import PadSeqUtil


class CL4KT(GDBaseModel):
    """
    CL4KT introduces a contrastive learning framework for knowledge tracing that learns effective representations by 
    pulling similar learning histories together and pushing dissimilar learning histories apart in representation space.
    """
    default_cfg = {
        'emb_size': 64,
        'hidden_size': 64,
        'num_blocks': 2,
        'num_attn_heads': 8,
        'kq_same': True,
        'dim_fc': 512,
        'd_ff': 1024,
        'l2': 0.0,
        'dropout': 0.2,
        'reg_cl': 0.1,
        'mask_prob': 0.2,
        'crop_prob': 0.3,
        'permute_prob': 0.3,
        'replace_prob': 0.3,
        'hard_negative': True,
        'temp': 0.05,
        'hard_negative_weight': 1.0,
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        """
        Get exercise nums and concepts nums.
        """
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']

    def build_model(self):
        """
        Initial model operator.
        """
        self.question_embed = nn.Embedding(
            self.n_cpt  + 2, self.modeltpl_cfg['emb_size']
        )
        self.interaction_embed = nn.Embedding(
            2 * (self.n_cpt + 2), self.modeltpl_cfg['emb_size']
        )
        self.sim = Similarity(temp=self.modeltpl_cfg["temp"])
        self.cpt_encoder = nn.ModuleList(
            [
                CL4KTTransformerLayer(
                    self.modeltpl_cfg['hidden_size'],
                    self.modeltpl_cfg['hidden_size'] // self.modeltpl_cfg['num_attn_heads'],
                    self.modeltpl_cfg['d_ff'],
                    self.modeltpl_cfg['num_attn_heads'],
                    self.modeltpl_cfg['dropout'],
                    self.modeltpl_cfg['kq_same']
                    )
                for _ in range(self.modeltpl_cfg['num_blocks'])
            ]
        )
        self.interaction_encoder = nn.ModuleList(
            [
                CL4KTTransformerLayer(
                    self.modeltpl_cfg['hidden_size'],
                    self.modeltpl_cfg['hidden_size'] // self.modeltpl_cfg['num_attn_heads'],
                    self.modeltpl_cfg['d_ff'],
                    self.modeltpl_cfg['num_attn_heads'],
                    self.modeltpl_cfg['dropout'],
                    self.modeltpl_cfg['kq_same']
                    )
                for _ in range(self.modeltpl_cfg['num_blocks'])
            ]
        )
        self.knoweldge_retriever = nn.ModuleList(
            [
                CL4KTTransformerLayer(
                    self.modeltpl_cfg['hidden_size'],
                    self.modeltpl_cfg['hidden_size'] // self.modeltpl_cfg['num_attn_heads'],
                    self.modeltpl_cfg['d_ff'],
                    self.modeltpl_cfg['num_attn_heads'],
                    self.modeltpl_cfg['dropout'],
                    self.modeltpl_cfg['kq_same']
                    )
                for _ in range(self.modeltpl_cfg['num_blocks'])
            ]
        )

        self.out = nn.Sequential(
            nn.Linear(self.modeltpl_cfg['hidden_size'] * 2, self.modeltpl_cfg['dim_fc']),
            nn.GELU(),
            nn.Dropout(self.modeltpl_cfg['dropout']),
            nn.Linear(self.modeltpl_cfg['dim_fc'], self.modeltpl_cfg['dim_fc'] // 2),
            nn.GELU(),
            nn.Dropout(self.modeltpl_cfg['dropout']),
            nn.Linear(self.modeltpl_cfg['dim_fc'] // 2, 1),
        )

    def forward(self, **kwargs):
        cpt_embed = self.question_embed(kwargs['cpt_unfold_seq'])
        inter_embed = self.get_interaction_embed(kwargs['cpt_unfold_seq'], kwargs['label_seq'].long(), kwargs['mask_seq'])

        x, y = cpt_embed, inter_embed
        for block in self.cpt_encoder:
            x, _ = block(mask=1, query=x, key=x, values=x, apply_pos=True) 
        for block in self.interaction_encoder:
            y, _ = block(mask=1, query=y, key=y, values=y, apply_pos=True)
        for block in self.knoweldge_retriever:
            x, att = block(mask=0, query=x, key=x, values=y, apply_pos=True)
        
        retrieved_knowledge = torch.cat([x, cpt_embed], dim=-1)
        # y_pd = torch.sigmoid(self.out(retrieved_knowledge)).squeeze()
        y_pd = torch.sigmoid(self.out(retrieved_knowledge))

        return y_pd, att
    
    def get_interaction_embed(self, cpts, responses, attention_mask):
        """
        Get interaction embedding by concepts and responses.
        """
        # masked_responses = responses * (attention_mask > 0).long()
        interactions = cpts + self.n_cpt * responses

        return self.interaction_embed(interactions)

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd,_ = self(**kwargs)
        y_pd = y_pd[:, 1:]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1].squeeze(dim=1)
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        """Loss calculation entry, including response prediciton loss and contrastive loss.

        Parameters
        ----------

            kwargs: dict
                kwargs is input feature embedding tensor including stu_id、exer_seq、cpt_unfold_seq、label_seq and mask_seq.
            stu_id: list
                Student id information.
            exer_seq: list
                Exercise is done by students.
            cpt_unfold_seq: list
                Concepts is done by students.
            label_seq: list
                It reprsents whether the exercise is done correctly.
            mask_seq: list
                Real data location.

        Returns
        -------
        torch.FloatTensor

            Response Prediction Loss and Contrastive Loss.
        """
        _, aug_c_seq_1, aug_r_seq_1, negative_r_seq, attention_mask_1 = self.augment_kt_seqs(**kwargs)
        _, aug_c_seq_2, aug_r_seq_2, _, attention_mask_2 = self.augment_kt_seqs(seed_change=True,**kwargs)
        aug_r_seq_1, aug_r_seq_2, negative_r_seq = aug_r_seq_1.long(), aug_r_seq_2.long(), negative_r_seq.long()
        attention_mask_1, attention_mask_2 = attention_mask_1.long(), attention_mask_2.long()

        # CL loss
        cpt_i_embed, cpt_j_embed = self.question_embed(aug_c_seq_1), self.question_embed(aug_c_seq_2)
        inter_i_embed, inter_j_embed = self.get_interaction_embed(aug_c_seq_1, aug_r_seq_1, attention_mask_1), self.get_interaction_embed(aug_c_seq_2, aug_r_seq_2, attention_mask_2)
        if self.modeltpl_cfg['hard_negative']: inter_k_embed = self.get_interaction_embed(kwargs['cpt_unfold_seq'], negative_r_seq, kwargs['mask_seq'])

        cpt_i_score, cpt_j_score = cpt_i_embed, cpt_j_embed
        inter_i_score, inter_j_score = inter_i_embed, inter_j_embed
        for block in self.cpt_encoder:
            cpt_i_score, _ = block(mask=2, query=cpt_i_score, key=cpt_i_score, values=cpt_i_embed, apply_pos=False)
            cpt_j_score, _ = block(mask=2, query=cpt_j_score, key=cpt_j_score, values=cpt_j_embed, apply_pos=False)
        for block in self.interaction_encoder:
            inter_i_score, _ = block(mask=2, query=inter_i_score, key=inter_i_score, values=inter_i_embed, apply_pos=False)
            inter_j_score, _ = block(mask=2, query=inter_j_score, key=inter_j_score, values=inter_j_embed, apply_pos=False)
            if self.modeltpl_cfg['hard_negative']: inter_k_embed, _ = block(mask=2, query=inter_k_embed, key=inter_k_embed, values=inter_k_embed, apply_pos=False)
        
        pool_cpt_i_score = (cpt_i_score * attention_mask_1.unsqueeze(-1)).sum(1) / attention_mask_1.sum(-1).unsqueeze(-1)
        pool_cpt_j_score = (cpt_j_score * attention_mask_2.unsqueeze(-1)).sum(1) / attention_mask_2.sum(-1).unsqueeze(-1)
        cpt_cos_sim = self.sim(pool_cpt_i_score.unsqueeze(1), pool_cpt_j_score.unsqueeze(0))
        cpt_labels = torch.arange(cpt_cos_sim.size(0)).long().to(aug_c_seq_1.device)
        cpt_cls_loss = F.cross_entropy(
            input=cpt_cos_sim, target=cpt_labels, reduction="mean"
        )

        pool_inter_i_score = (inter_i_score * attention_mask_1.unsqueeze(-1)).sum(1) / attention_mask_1.sum(-1).unsqueeze(-1)
        pool_inter_j_score = (inter_j_score * attention_mask_2.unsqueeze(-1)).sum(1) / attention_mask_2.sum(-1).unsqueeze(-1)
        inter_cos_sim = self.sim(pool_inter_i_score.unsqueeze(1), pool_inter_j_score.unsqueeze(0))
        if self.modeltpl_cfg['hard_negative']:
            pool_inter_k_score = (inter_k_embed * kwargs['mask_seq'].unsqueeze(-1)).sum(1) / kwargs['mask_seq'].sum(-1).unsqueeze(-1)
            neg_inter_cos_sim = self.sim(pool_inter_i_score.unsqueeze(1), pool_inter_k_score.unsqueeze(0))
            inter_cos_sim = torch.cat([inter_cos_sim, neg_inter_cos_sim], 1)
        inter_labels = torch.arange(inter_cos_sim.size(0)).long().to(aug_c_seq_1.device)
        if self.modeltpl_cfg['hard_negative']:
            weights = torch.tensor(
                    [
                        [0.0] * (inter_cos_sim.size(-1) - neg_inter_cos_sim.size(-1))
                        + [0.0] * i
                        + [self.modeltpl_cfg['hard_negative_weight']]
                        + [0.0] * (neg_inter_cos_sim.size(-1) - i - 1)
                        for i in range(neg_inter_cos_sim.size(-1))
                    ]
            ).to(aug_c_seq_1.device)
            inter_cos_sim = inter_cos_sim + weights
        inter_cls_loss = F.cross_entropy(
            input=inter_cos_sim, target=inter_labels, reduction="mean"
        )
        cl_loss = cpt_cls_loss + inter_cls_loss

        # Predict
        y_pd, _ = self(**kwargs)

        y_pd = y_pd[:, 1:]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1].squeeze(dim=1)
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt, reduction="mean"
        )
        return {
            'loss_main': loss + cl_loss * self.modeltpl_cfg['reg_cl']
        }
        

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)

    def augment_kt_seqs(self, seed_change=False, **kwargs):
        """ Get augmental data, including four methods: Question mask、Interaction crop、Interaction permute、Question replac.

        * **Question mask**: replace some questions in the original history with a special token [mask], without changing their responses.
        * **Interaction crop**: create a random subset of original data.
        * **Interaction permute**: re-order interactions in a sub-sequence of the original history.
        * **Question replac**: convert original questions to easier or more difficult questions based on their responses.

        Returns: 
            tuple:
                - exer_seq_final: augmental exercise sequence.
                - cpt_seq_fianl: augmental concept sequence.
                - label_seq_final: new label after augment.
                - label_seq_flip: flipping the label for hard negative.
                - mask: new mask matrix after augment.
        """
        if seed_change:
            random.Random(self.traintpl_cfg['seed'] + 1)
            np.random.seed(self.traintpl_cfg['seed'] + 1)
        bs = kwargs['exer_seq'].size(0)
        lens = (kwargs['mask_seq'] > 0).sum(dim=1)

        # Data Augmentation
        cpt_seq_ = kwargs['cpt_unfold_seq'].clone()
        label_seq_ = kwargs['label_seq'].clone()
        exer_seq_ = kwargs['exer_seq'].clone()
        
        # Manipulate order: Question mask
        if self.modeltpl_cfg['mask_prob'] > 0:
           for b in range(bs):
                if self.datatpl_cfg['M2C_CL4KT_OP']['sequence_truncation'] == 'recent':
                    idx = random.sample(range(self.datatpl_cfg['M2C_CL4KT_OP']['window_size']-lens[b], self.datatpl_cfg['M2C_CL4KT_OP']['window_size']), max(1, int(lens[b] * self.modeltpl_cfg['mask_prob'])))
                else:
                    idx = random.sample(range(lens[b]-1), max(1, int(lens[b] * self.modeltpl_cfg['mask_prob'])))
                for i in idx:
                        cpt_seq_[b, i] = self.n_cpt + 1
                        exer_seq_[b, i] = self.n_item + 1
        # Hard negative
        label_seq_flip = kwargs['label_seq'].clone() if self.modeltpl_cfg['hard_negative'] else label_seq_
        for b in range(bs):
            if self.datatpl_cfg['M2C_CL4KT_OP']['sequence_truncation'] == 'recent':
                idx = torch.arange(self.datatpl_cfg['M2C_CL4KT_OP']['window_size']-lens[b], self.datatpl_cfg['M2C_CL4KT_OP']['window_size'])
            else:
                idx = torch.arange(lens[b])
            for i in idx:
                label_seq_flip[b, i] = 1 - label_seq_flip[b, i]
        # Manipulate order:Question replace
        if self.modeltpl_cfg['replace_prob'] > 0:
            for b in range(bs):
                if self.datatpl_cfg['M2C_CL4KT_OP']['sequence_truncation'] == 'recent':
                    idx = random.sample(range(self.datatpl_cfg['M2C_CL4KT_OP']['window_size']-lens[b], self.datatpl_cfg['M2C_CL4KT_OP']['window_size']), max(1, int(lens[b] * self.modeltpl_cfg['replace_prob'])))
                else:
                    idx = random.sample(range(lens[b]-1), max(1, int(lens[b] * self.modeltpl_cfg['replace_prob'])))
                for i in idx:
                    if cpt_seq_[b, i] != self.n_cpt + 1 and i in self.datatpl_cfg['dt_info']['train_harder_cpts'] and i in self.datatpl_cfg['dt_info']['train_harder_cpts']:
                        cpt_seq_[b, i] = self.datatpl_cfg['dt_info']['train_harder_cpts'][i]  if label_seq_[b, i] == 0 else self.datatpl_cfg['dt_info']['train_harder_cpts'][i]
        # Manipulate order:Interaction permute
        if self.modeltpl_cfg['permute_prob'] > 0:
            for b in range(bs):
                reorder_seq_len = int(lens[b] * self.modeltpl_cfg['permute_prob'])
                if self.datatpl_cfg['M2C_CL4KT_OP']['sequence_truncation'] == 'recent':
                    start_pos = random.sample(range(self.datatpl_cfg['M2C_CL4KT_OP']['window_size']-lens[b], self.datatpl_cfg['M2C_CL4KT_OP']['window_size']-reorder_seq_len), 1)
                else:
                    start_pos = random.sample(range(lens[b]-reorder_seq_len-1), 1)

                perm = np.random.permutation(reorder_seq_len)
                exer_seq_[b, start_pos[0]:start_pos[0]+reorder_seq_len] = exer_seq_[b, start_pos[0]:start_pos[0]+reorder_seq_len][perm]
                cpt_seq_[b, start_pos[0]:start_pos[0]+reorder_seq_len] = cpt_seq_[b, start_pos[0]:start_pos[0]+reorder_seq_len][perm]
                label_seq_[b, start_pos[0]:start_pos[0]+reorder_seq_len] = label_seq_[b, start_pos[0]:start_pos[0]+reorder_seq_len][perm]
        # Manipulate order:Interaction crop
        exer_seq_final, cpt_seq_fianl, label_seq_final, mask = torch.zeros_like(exer_seq_), torch.zeros_like(cpt_seq_), torch.zeros_like(label_seq_), torch.zeros_like(kwargs['mask_seq'])
        if self.modeltpl_cfg['crop_prob'] > 0:
            for b in range(bs):
                crop_seq_len = 1 if int(lens[b] * self.modeltpl_cfg['permute_prob']) == 0 else int(lens[b] * self.modeltpl_cfg['permute_prob'])
                if self.datatpl_cfg['M2C_CL4KT_OP']['sequence_truncation'] == 'recent':
                    start_pos = random.sample(range(self.datatpl_cfg['M2C_CL4KT_OP']['window_size']-lens[b], self.datatpl_cfg['M2C_CL4KT_OP']['window_size']-crop_seq_len + 1), 1)
                    exer_seq_final[b, -crop_seq_len:] = exer_seq_[b, start_pos[0]:start_pos[0] + crop_seq_len]
                    cpt_seq_fianl[b, -crop_seq_len:] = cpt_seq_[b, start_pos[0]:start_pos[0] + crop_seq_len]
                    label_seq_final[b, -crop_seq_len:] = label_seq_[b, start_pos[0]:start_pos[0] + crop_seq_len]
                    mask[b, -crop_seq_len:] = 1
                else:
                    start_pos = random.sample(range(self.datatpl_cfg['M2C_CL4KT_OP']['window_size']-lens[b], self.datatpl_cfg['M2C_CL4KT_OP']['window_size']-crop_seq_len-1), 1)
                    exer_seq_final[b, :crop_seq_len] = exer_seq_[b, start_pos[0]:start_pos[0] + crop_seq_len]
                    cpt_seq_fianl[b, :crop_seq_len] = cpt_seq_[b, start_pos[0]:start_pos[0] + crop_seq_len]
                    label_seq_final[b, :crop_seq_len] = label_seq_[b, start_pos[0]:start_pos[0] + crop_seq_len]
                    mask[b, crop_seq_len] = 1

        return exer_seq_final, cpt_seq_fianl, label_seq_final, label_seq_flip, mask

class Similarity(nn.Module):
    """ Calculate cosine similarity between data x and data y.
    """
    def __init__(self, temp):
        super().__init__()
        self.temp = temp
        self.cos = nn.CosineSimilarity(dim=-1)

    def forward(self, x, y):
        return self.cos(x, y) / self.temp


class CL4KTTransformerLayer(nn.Module):
    """
    This is a Basic Block of Transformer paper. It contains one Multi-head attention object.

    Followed by layer norm and position-wise feed-forward net and dropotu layer.
    """
    def __init__(self, d_model, d_feature, d_ff, n_heads, dropout, kq_same):
        super(CL4KTTransformerLayer, self).__init__()
        # Multi-Head Attention Block
        self.masked_attn_head = MultiHeadAttentionWithIndividualFeatures(
            d_model, d_feature, n_heads, dropout, kq_same=kq_same
        )
        # Two layer norm and two dropout layers
        self.layer_norm1 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)

        self.linear1 = nn.Linear(d_model, d_ff)
        self.activation = nn.GELU()
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(d_ff, d_model)
        self.layer_norm2 = nn.LayerNorm(d_model)
        self.dropout2 = nn.Dropout(dropout)
        

    def forward(self, mask, query, key, values, apply_pos=True):
        # device = query.get_device()
        seqlen =  query.size(1)

        nopeek_mask = np.triu(np.ones((1, 1, seqlen, seqlen)), k=mask)
        src_mask = (torch.from_numpy(nopeek_mask) == 0).to(query.device)
        bert_mask = torch.ones_like(src_mask).bool()

        if mask == 0:
            query2, attn = self.masked_attn_head(query, key, values, mask=src_mask)
        elif mask == 1:
            query2, attn = self.masked_attn_head(query, key, values, mask=src_mask)
        else:  # mask == 2
            query2, attn = self.masked_attn_head(query, key, values, mask=bert_mask)
        
        query = query + self.dropout1((query2))  # residual connection
        query = self.layer_norm1(query)

        if apply_pos:
            query2 = self.linear2(self.dropout(self.activation(self.linear1(query))))
            query = query + self.dropout2((query2))
            query = self.layer_norm2(query)

        return query, attn


class MultiHeadAttentionWithIndividualFeatures(nn.Module):
    """
    It has projection layer for getting keys, queries, and values. 
    
    Followed by attention and a connected layer.
    """
    def __init__(self, d_model, d_feature, n_heads, dropout, kq_same, bias=True):
        super(MultiHeadAttentionWithIndividualFeatures, self).__init__()
        self.d_model = d_model
        self.d_k = d_feature
        self.h = n_heads
        self.kq_same = kq_same

        self.v_linear = nn.Linear(d_model, d_model, bias=bias)
        self.k_linear = nn.Linear(d_model, d_model, bias=bias)
        if kq_same is False:
            self.q_linear = nn.Linear(d_model, d_model, bias=bias)
        self.dropout = nn.Dropout(dropout)
        self.proj_bias = bias
        self.out_proj = nn.Linear(d_model, d_model, bias=bias)
        self.gammas = nn.Parameter(torch.zeros(n_heads, 1, 1))
        torch.nn.init.xavier_uniform_(self.gammas)

        self._reset_parameters()
    
    def _reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.k_linear.weight)
        torch.nn.init.xavier_uniform_(self.v_linear.weight)
        if self.kq_same is False:
            torch.nn.init.xavier_uniform_(self.q_linear.weight)

        if self.proj_bias:
            torch.nn.init.constant_(self.k_linear.bias, 0.0)
            torch.nn.init.constant_(self.v_linear.bias, 0.0)
            if self.kq_same is False:
                torch.nn.init.constant_(self.q_linear.bias, 0.0)
            torch.nn.init.constant_(self.out_proj.bias, 0.0)
    
    def forward(self, q, k, v, mask):
        bs = q.size(0)

        # perform linear operation and split into h heads
        k = self.k_linear(k).view(bs, -1, self.h, self.d_k)
        if self.kq_same is False:
            q = self.q_linear(q).view(bs, -1, self.h, self.d_k)
        else:
            q = self.k_linear(q).view(bs, -1, self.h, self.d_k)
        v = self.v_linear(v).view(bs, -1, self.h, self.d_k)

        # transpose to get dimensions batch_size * num_heads * seqlen * d_model
        k = k.transpose(1, 2)
        q = q.transpose(1, 2)
        v = v.transpose(1, 2)

        # calculate attention using function we will define next
        gammas = self.gammas
        scores, attn_scores = individual_attention(
            q, k, v, self.d_k, mask, self.dropout, gammas
        )

        # concatenate heads and put through final linear layer
        concat = scores.transpose(1, 2).contiguous().view(bs, -1, self.d_model)

        # concat torch.Size([24, 200, 256])   [batch_size, seqlen, d_model]
        output = self.out_proj(concat)

        return output, attn_scores

def individual_attention(q, k, v, d_k, mask, dropout, gamma=None):
    """
    This is called by MultiHeadAttention object to find the values.
    """
    scores = torch.matmul(q, k.transpose(-2, -1)) / np.sqrt(
        d_k
    )  # [batch_size, 8, seq_len, seq_len]
    bs, head, seqlen = scores.size(0), scores.size(1), scores.size(2)

    x1 = torch.arange(seqlen).expand(seqlen, -1)
    x2 = x1.transpose(0, 1).contiguous()

    with torch.no_grad():
        scores_ = scores.masked_fill(mask == 0, -1e32)
        scores_ = F.softmax(scores_, dim=-1)
        scores_ = scores_ * mask.float()

        distcum_scores = torch.cumsum(scores_, dim=-1)

        disttotal_scores = torch.sum(scores_, dim=-1, keepdim=True)

        # device = distcum_scores.get_device()
        position_effect = torch.abs(x1 - x2)[None, None, :, :].type(torch.FloatTensor)
        position_effect = position_effect.to(distcum_scores.device)

        dist_scores = torch.clamp(
            (disttotal_scores - distcum_scores) * position_effect, min=0.0
        )
        dist_scores = dist_scores.sqrt().detach()

    m = nn.Softplus()

    gamma = -1.0 * m(gamma).unsqueeze(0)

    total_effect = torch.clamp(
        torch.clamp((dist_scores * gamma).exp(), min=1e-5), max=1e5
    )

    scores = scores * total_effect

    scores.masked_fill_(mask == 0, -1e32)
    scores = F.softmax(scores, dim=-1)

    attn_scores = scores
    scores = dropout(scores)
    output = torch.matmul(scores, v)
    return output, attn_scores



from .base_evaltpl import BaseEvalTPL
import torch
import pandas as pd
import numpy as np
from edustudio.utils.common import tensor2npy
from edustudio.utils.callback import ModeState
from tqdm import tqdm


class IdentifiabilityEvalTPL(BaseEvalTPL):
    default_cfg = {
        'use_metrics': ['IDS'], # Identifiability Score
        'test_only_metrics': ['IDS']
    }

    def eval(self, stu_stats:np.ndarray,  **kwargs):
        metric_result = {}
        ignore_metrics = kwargs.get('ignore_metrics', {})
        for metric_name in self.evaltpl_cfg[self.name]['use_metrics']:
            if metric_name not in ignore_metrics:
                if metric_name in self.evaltpl_cfg[self.name]['test_only_metrics'] and \
                    self.callback_list.mode_state != ModeState.END:
                   continue
                if metric_name == "IDS":
                    metric_result[metric_name] = self._get_metrics(metric_name)(
                       self.log_mat_total, stu_stats
                    )
                else:
                    raise ValueError(f"unknown metric_name: {metric_name}")
        return metric_result

    def _get_metrics(self, metric):
        if metric == "IDS":
            return self.get_IDS
        else:
            raise NotImplementedError
    
    def set_dataloaders(self, train_loader, test_loader, valid_loader=None):
        super().set_dataloaders(train_loader, test_loader, valid_loader)

        if self.valid_loader:
            
            self.stu_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['stu_id'],
                self.valid_loader.dataset.dict_main['stu_id'],
                self.test_loader.dataset.dict_main['stu_id'],
            ]))
            self.exer_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['exer_id'],
                self.valid_loader.dataset.dict_main['exer_id'],
                self.test_loader.dataset.dict_main['exer_id'],
            ]))

            self.label_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['label'],
                self.valid_loader.dataset.dict_main['label'],
                self.test_loader.dataset.dict_main['label'],
            ]))

        else:
            self.stu_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['stu_id'],
                self.test_loader.dataset.dict_main['stu_id'],
            ]))
            self.exer_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['exer_id'],
                self.test_loader.dataset.dict_main['exer_id'],
            ]))

            self.label_id_total = tensor2npy(torch.cat([
                self.train_loader.dataset.dict_main['label'],
                self.test_loader.dataset.dict_main['label'],
            ]))

        self.log_mat_total = self._gen_log_mat(self.stu_id_total, self.exer_id_total, self.label_id_total)
  
    def _gen_log_mat(self, uid, iid, label):
        n_stu = self.datatpl_cfg['dt_info']['stu_count']
        n_exer = self.datatpl_cfg['dt_info']['exer_count']
        label = label.copy()
        label[label == 0] = -1
        log_mat = np.zeros((n_stu, n_exer), dtype=np.float32)
        log_mat[uid, iid] = label
        return log_mat

    def get_IDS(self, R, T):
        """
        R: interaction matrix
        T: student proficiency matrix
        """
        count = 0
        val = 0.0
        n_stu = R.shape[0]
        assert R.shape[0] == T.shape[0]
        for s_i in range(n_stu):
            for s_j in range(s_i+1, n_stu):
                if not np.all(R[s_i] == R[s_j]):
                    continue
                count += 1
                val += (1+np.abs(T[s_i] - T[s_j]).sum())**2
        if val == 0:
            self.logger.error(f"invalid computation occurred in {self.name}")
            return 0.0
        return (count / val).item()



r"""
KaNCD
##########################################

Reference:
    Fei Wang et al. "NeuralCD: A General Framework for Cognitive Diagnosis" in TKDE 2022.

Reference Code:
    https://github.com/bigdata-ustc/EduCDM/tree/main/EduCDM/KaNCD

"""

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
from ..utils.components import PosLinear
import torch
import torch.nn.functional as F



class KaNCD(GDBaseModel):
    default_cfg = {
        'prednet_len1': 256,
        'prednet_len2': 128,
        'emb_dim': 20,
        'mf_type': 'gmf'
    }
    def __init__(self, cfg):
        super().__init__(cfg)
        assert self.modeltpl_cfg['mf_type'] in ['mf', 'gmf', 'ncf1', 'ncf2']

    def add_extra_data(self, **kwargs):
        self.Q_mat = kwargs['Q_mat'].to(self.device)

    def build_cfg(self):
        self.knowledge_n = self.datatpl_cfg['dt_info']['cpt_count']
        self.exer_n = self.datatpl_cfg['dt_info']['exer_count']
        self.student_n = self.datatpl_cfg['dt_info']['stu_count']
        self.emb_dim = self.modeltpl_cfg['emb_dim']
        self.mf_type = self.modeltpl_cfg['mf_type']
        self.prednet_input_len = self.knowledge_n
        self.prednet_len1, self.prednet_len2 = self.modeltpl_cfg['prednet_len1'], self.modeltpl_cfg['prednet_len2']

    def build_model(self):
        # prediction sub-net
        self.student_emb = nn.Embedding(self.student_n, self.emb_dim)
        self.exercise_emb = nn.Embedding(self.exer_n, self.emb_dim)
        self.knowledge_emb = nn.Parameter(torch.zeros(self.knowledge_n, self.emb_dim))
        self.e_discrimination = nn.Embedding(self.exer_n, 1)
        self.prednet_full1 = PosLinear(self.prednet_input_len, self.prednet_len1)
        self.drop_1 = nn.Dropout(p=0.5)
        self.prednet_full2 = PosLinear(self.prednet_len1, self.prednet_len2)
        self.drop_2 = nn.Dropout(p=0.5)
        self.prednet_full3 = PosLinear(self.prednet_len2, 1)

        if self.mf_type == 'gmf':
            self.k_diff_full = nn.Linear(self.emb_dim, 1)
            self.stat_full = nn.Linear(self.emb_dim, 1)
        elif self.mf_type == 'ncf1':
            self.k_diff_full = nn.Linear(2 * self.emb_dim, 1)
            self.stat_full = nn.Linear(2 * self.emb_dim, 1)
        elif self.mf_type == 'ncf2':
            self.k_diff_full1 = nn.Linear(2 * self.emb_dim, self.emb_dim)
            self.k_diff_full2 = nn.Linear(self.emb_dim, 1)
            self.stat_full1 = nn.Linear(2 * self.emb_dim, self.emb_dim)
            self.stat_full2 = nn.Linear(self.emb_dim, 1)

        nn.init.xavier_normal_(self.knowledge_emb)

    def forward(self, stu_id, exer_id, **kwargs):
        input_exercise = exer_id
        input_knowledge_point = self.Q_mat[exer_id]
        # before prednet
        stu_emb = self.student_emb(stu_id)
        exer_emb = self.exercise_emb(input_exercise)
        # get knowledge proficiency
        batch, dim = stu_emb.size()
        stu_emb = stu_emb.view(batch, 1, dim).repeat(1, self.knowledge_n, 1)
        knowledge_emb = self.knowledge_emb.repeat(batch, 1).view(batch, self.knowledge_n, -1)
        if self.mf_type == 'mf':  # simply inner product
            stat_emb = torch.sigmoid((stu_emb * knowledge_emb).sum(dim=-1, keepdim=False))  # batch, knowledge_n
        elif self.mf_type == 'gmf':
            stat_emb = torch.sigmoid(self.stat_full(stu_emb * knowledge_emb)).view(batch, -1)
        elif self.mf_type == 'ncf1':
            stat_emb = torch.sigmoid(self.stat_full(torch.cat((stu_emb, knowledge_emb), dim=-1))).view(batch, -1)
        elif self.mf_type == 'ncf2':
            stat_emb = torch.sigmoid(self.stat_full1(torch.cat((stu_emb, knowledge_emb), dim=-1)))
            stat_emb = torch.sigmoid(self.stat_full2(stat_emb)).view(batch, -1)
        batch, dim = exer_emb.size()
        exer_emb = exer_emb.view(batch, 1, dim).repeat(1, self.knowledge_n, 1)
        if self.mf_type == 'mf':
            k_difficulty = torch.sigmoid((exer_emb * knowledge_emb).sum(dim=-1, keepdim=False))  # batch, knowledge_n
        elif self.mf_type == 'gmf':
            k_difficulty = torch.sigmoid(self.k_diff_full(exer_emb * knowledge_emb)).view(batch, -1)
        elif self.mf_type == 'ncf1':
            k_difficulty = torch.sigmoid(self.k_diff_full(torch.cat((exer_emb, knowledge_emb), dim=-1))).view(batch, -1)
        elif self.mf_type == 'ncf2':
            k_difficulty = torch.sigmoid(self.k_diff_full1(torch.cat((exer_emb, knowledge_emb), dim=-1)))
            k_difficulty = torch.sigmoid(self.k_diff_full2(k_difficulty)).view(batch, -1)
        # get exercise discrimination
        e_discrimination = torch.sigmoid(self.e_discrimination(input_exercise))

        # prednet
        input_x = e_discrimination * (stat_emb - k_difficulty) * input_knowledge_point
        # f = input_x[input_knowledge_point == 1]
        input_x = self.drop_1(torch.tanh(self.prednet_full1(input_x)))
        input_x = self.drop_2(torch.tanh(self.prednet_full2(input_x)))
        output_1 = torch.sigmoid(self.prednet_full3(input_x))

        return output_1.view(-1)
    
    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    @torch.no_grad()
    def predict(self, stu_id, exer_id,**kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }

    def get_stu_status(self, stu_id=None):
        if stu_id is not None:
            stu_emb = self.student_emb(stu_id)
        else:
            stu_emb = self.student_emb.weight
        batch, dim = stu_emb.size()
        stu_emb = stu_emb.view(batch, 1, dim).repeat(1, self.knowledge_n, 1)
        knowledge_emb = self.knowledge_emb.repeat(batch, 1).view(batch, self.knowledge_n, -1)
        if self.mf_type == 'mf':  # simply inner product
            stat_emb = torch.sigmoid((stu_emb * knowledge_emb).sum(dim=-1, keepdim=False))  # batch, knowledge_n
        elif self.mf_type == 'gmf':
            stat_emb = torch.sigmoid(self.stat_full(stu_emb * knowledge_emb)).view(batch, -1)
        elif self.mf_type == 'ncf1':
            stat_emb = torch.sigmoid(self.stat_full(torch.cat((stu_emb, knowledge_emb), dim=-1))).view(batch, -1)
        elif self.mf_type == 'ncf2':
            stat_emb = torch.sigmoid(self.stat_full1(torch.cat((stu_emb, knowledge_emb), dim=-1)))
            stat_emb = torch.sigmoid(self.stat_full2(stat_emb)).view(batch, -1)
        return stat_emb



from .base_evaltpl import BaseEvalTPL
import pandas as pd
import numpy as np
from edustudio.utils.common import tensor2npy


class FairnessEvalTPL(BaseEvalTPL):
    """Fairness Cognitive Evaluation
    """
    default_cfg = {
        'use_sensi_attrs': ['gender:token'],
        'use_metrics': ['EO', 'DP', 'FCD']
    }

    def _check_params(self):
        assert len(set(self.evaltpl_cfg[self.__class__.__name__]['use_metrics']) - {'EO', 'DP', 'FCD'}) == 0

    def eval(self, **kwargs):
        stu_id = tensor2npy(kwargs['stu_id'])
        pd_soft = tensor2npy(kwargs['y_pd'])
        gt = tensor2npy(kwargs['y_gt'])
        pd_hard = (pd_soft >= 0.5).astype(np.int64)

        df_stu = self.extra_data['df_stu']

        df = pd.DataFrame()
        df['stu_id:token'] = stu_id
        df['pd_soft'] = pd_soft
        df['pd_hard'] = pd_hard
        df['gt'] = gt
        df = df.merge(df_stu, on='stu_id:token', how='left')
        
        ret_dic = {}
        for attr in self.evaltpl_cfg[self.__class__.__name__]['use_sensi_attrs']:
            g_names = df_stu[attr].unique()

            for use_metric in self.evaltpl_cfg[self.__class__.__name__]['use_metrics']:
                if len(g_names) == 2:
                    if use_metric == 'EO': ret_dic[f"EO_{attr}"] = self.get_eo(df, attr)
                    if use_metric == 'DP': ret_dic[f"DP_{attr}"] = self.get_dp(df, attr)
                    if use_metric == 'FCD': ret_dic[f"FCD_{attr}"] = self.get_fcd(df, attr)
                else:
                    pass
        return ret_dic

    def get_dp(self, df, sensitive_attr):
        """Demographic Parity
        """
        dp = df.groupby(sensitive_attr)['pd_hard'].mean()
        return abs(dp[0] - dp[1])

    def get_eo(self, df, sensitive_attr):
        """Equal Opportunity
        """
        eo = df.groupby([sensitive_attr, 'gt'])['pd_hard'].mean()
        return abs(eo[0][1] - eo[1][1])


    def get_fcd(self, df, sensitive_attr):
        """Fair Cognitive Diagnosis [1]
        [1]zhang zheng, et al, Understanding and Improving Fairness in Cognitive Diagnosis,  SCIENCE CHINA Information Sciences, 2023, ISSN 1674-733X, https://doi.org/10.1007/s11432-022-3852-0.
        """
        fcd_pd = df.groupby([sensitive_attr, 'stu_id:token'])['pd_hard'].mean()
        fcd_pd = fcd_pd[0].mean() - fcd_pd[1].mean()

        fcd_gt = df.groupby([sensitive_attr, 'stu_id:token'])['gt'].mean()
        fcd_gt = fcd_gt[0].mean() - fcd_gt[1].mean()
        return abs(fcd_pd - fcd_gt)


    def add_extra_data(self, **kwargs):
        self.extra_data = kwargs

        df_stu = self.extra_data['df_stu']
        assert df_stu is not None
        for attr in self.evaltpl_cfg[self.__class__.__name__]['use_sensi_attrs']:
            assert attr in df_stu
            g_names = df_stu[attr].unique()

            for use_metric in self.evaltpl_cfg[self.__class__.__name__]['use_metrics']:
                assert len(g_names) >= 2
                if len(g_names) > 2:
                    if use_metric in {'EO', 'DP', 'FCD'}:
                        self.logger.warning(f"As the number of sensitive attribute `{attr}` values > 2, the fairness metric {use_metric} is not supported for the `{attr}`")



from ..common import EduDataTPL

class FAIRDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_FilteringRecordsByAttr', 'M2C_FilterRecords4CD', 'M2C_ReMapId', 'M2C_RandomDataSplit4CD', 'M2C_GenQMat'],
    }




from .irt import IRT
from .mf import MF
from .mirt import MIRT
from .ncdm import NCDM
from .dina import DINA, STEDINA
from .kancd import KaNCD
from .kscd import KSCD
from .cncd_q import CNCD_Q
from .irr import IRR
from .hier_cdf import HierCDF
from .rcd import RCD
from .cncd_f import CNCD_F
from .cdgk import CDGK_SINGLE, CDGK_MULTI
from .cdmfkc import CDMFKC
from .ecd import *
from .mgcd import MGCD
from .dcd import DCD
from .faircd import FairCD_IRT, FairCD_MIRT, FairCD_NCDM



from ..common import BaseMid2Cache
import numpy as np
import importlib


class M2C_RCD_OP(BaseMid2Cache):
    default_cfg = {}

    def process(self, **kwargs):        
        self.dgl = importlib.import_module("dgl")
        k_e, e_k = self.build_k_e(kwargs)
        u_e_list, e_u_list = self.build_u_e(kwargs)
        local_map = {
            'directed_g': self.build_cpt_directd(kwargs),
            'undirected_g': self.build_cpt_undirected(kwargs),
            'k_from_e': k_e,
            'e_from_k': e_k,
            'u_from_e_list': u_e_list,
            'e_from_u_list': e_u_list,
        }
        kwargs['local_map'] = local_map
        return kwargs

    def build_cpt_undirected(self, kwargs):
        cpt_count =  kwargs['dt_info']['cpt_count']
        cpt_dep_mat = kwargs['cpt_dep_mat']
        cpt_dep_mat_undirect = ((cpt_dep_mat + cpt_dep_mat.T) == 2).astype(np.int64)
        # undirected (only prerequisite)
        g_undirected = self.dgl.graph(np.argwhere(cpt_dep_mat_undirect == 1).tolist(), num_nodes=cpt_count)
        return g_undirected
    
    def build_cpt_directd(self, kwargs):
        cpt_count =  kwargs['dt_info']['cpt_count']
        cpt_dep_mat = kwargs['cpt_dep_mat']
        # directed (prerequisite + similarity)
        g_directed =self.dgl.graph(np.argwhere(cpt_dep_mat == 1).tolist(), num_nodes=cpt_count)
        return g_directed
    
    def build_k_e(self, kwargs):
        cpt_count = kwargs['dt_info']['cpt_count']
        exer_count = kwargs['dt_info']['exer_count']
        df_exer = kwargs['df_exer']
        
        edges = df_exer[['exer_id:token','cpt_seq:token_seq']].explode('cpt_seq:token_seq').to_numpy()
        edges[:, 1] += exer_count

        k_e = self.dgl.graph(edges.tolist(), num_nodes=cpt_count + exer_count)
        e_k = self.dgl.graph(edges[:,[1,0]].tolist(), num_nodes=cpt_count + exer_count)
        return k_e, e_k

    def build_u_e(self, kwargs):
        stu_count = kwargs['dt_info']['stu_count']
        exer_count = kwargs['dt_info']['exer_count']
        df_train_folds = kwargs['df_train_folds']

        u_from_e_list= []
        e_from_u_list = []
        for train_df in df_train_folds:
            stu_id = train_df['stu_id:token'].to_numpy() + exer_count
            exer_id = train_df['exer_id:token'].to_numpy()
            u_e = self.dgl.graph(np.vstack([exer_id, stu_id]).T.tolist(), num_nodes=stu_count + exer_count)
            e_u = self.dgl.graph(np.vstack([stu_id, exer_id]).T.tolist(), num_nodes=stu_count + exer_count)
            u_from_e_list.append(u_e)
            e_from_u_list.append(e_u)
        return u_from_e_list, e_from_u_list



from ..gd_basemodel import GDBaseModel
import random
import math
import torch.nn as nn
import torch
import torch.nn.functional as F

MIN_SEQ_LEN = 5

class DTransformer(GDBaseModel):
    default_cfg = {
        'n_knowledges': 16,
        'hidden_size': 128,
        'dim_fc': 256,
        'num_layers': 3,
        'num_heads': 8,
        'lambda_cl': 0.1,
        'prediction_window': 1,
        'dropout': 0.2,
        'projection_alhead_cl': 0,
        'cl_loss': True,
        'hard_negative': True,
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_exer = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']

    def build_model(self):
        self.cpt_embed = nn.Embedding(self.n_cpt + 1, self.modeltpl_cfg['hidden_size'])
        self.label_embed = nn.Embedding(2, self.modeltpl_cfg['hidden_size'])
        
        self.cpt_diff_embed = nn.Embedding(self.n_cpt + 1, self.modeltpl_cfg['hidden_size'])
        self.label_diff_embed = nn.Embedding(2, self.modeltpl_cfg['hidden_size'])
        self.exer_diff_embed = nn.Embedding(self.n_exer + 1, 1)  # difficult parameter

        self.block1 = DTransformerLayer(self.modeltpl_cfg['hidden_size'], self.modeltpl_cfg['num_heads'], self.modeltpl_cfg['dropout'])
        self.block2 = DTransformerLayer(self.modeltpl_cfg['hidden_size'], self.modeltpl_cfg['num_heads'], self.modeltpl_cfg['dropout'], kq_same=False)

        self.know_params = nn.Parameter(torch.empty(self.modeltpl_cfg['n_knowledges'], self.modeltpl_cfg['hidden_size']))
        nn.init.uniform_(self.know_params, -1.0, 1.0)

        self.out = nn.Sequential(
            nn.Linear(self.modeltpl_cfg['hidden_size'] * 2, self.modeltpl_cfg['dim_fc']),
            nn.GELU(),
            nn.Dropout(self.modeltpl_cfg['dropout']),
            nn.Linear(self.modeltpl_cfg['dim_fc'], self.modeltpl_cfg['dim_fc'] // 2),
            nn.GELU(),
            nn.Dropout(self.modeltpl_cfg['dropout']),
            nn.Linear(self.modeltpl_cfg['dim_fc'] // 2, 1),
        )

        if self.modeltpl_cfg['projection_alhead_cl']:
            self.proj = nn.Sequential(nn.Linear(self.modeltpl_cfg['hidden_size'], self.modeltpl_cfg['hidden_size']), nn.GELU())
        else:
            self.proj = None
        
    def forward(self, exer_seq, cpt_unfold_seq, label_seq, mask_seq, is_train=True, n=1,**kwargs):
        lens = (mask_seq > 0).sum(dim=1)
        q_emb, a_emb, exer_diff = self.embedding(cpt_unfold_seq, label_seq, exer_seq)
        
        if self.modeltpl_cfg['num_layers'] == 1:
            hq = q_emb.clone()
            ha = a_emb.clone()
            m, e_score = self.block1(hq, hq, ha, lens, peek_cur=True)
        elif self.modeltpl_cfg['num_layers'] == 2:
            hq = q_emb.clone()
            ha, _ = self.block1(a_emb, a_emb, a_emb, lens, peek_cur=True)
            m, e_score = self.block1(hq, hq, ha, lens, peek_cur=True)
        else:
            hq, _ = self.block1(q_emb, q_emb, q_emb, lens, peek_cur=True)
            ha, _ = self.block1(a_emb, a_emb, a_emb, lens, peek_cur=True)
            m, e_score = self.block1(hq, hq, ha, lens, peek_cur=True)

        bs, seqlen = m.size(0), m.size(1)

        query = (
            self.know_params[None, :, None, :]
            .expand(bs, -1, seqlen, -1)
            .contiguous()
            .view(bs * self.modeltpl_cfg['n_knowledges'], seqlen, self.modeltpl_cfg['hidden_size'])
        )
        hq = hq.unsqueeze(1).expand(-1, self.modeltpl_cfg['n_knowledges'], -1, -1).reshape_as(query)  # key embedding
        m = m.unsqueeze(1).expand(-1, self.modeltpl_cfg['n_knowledges'], -1, -1).reshape_as(query)  # value embedding
        z, c_score = self.block2(query, hq, m, torch.repeat_interleave(lens, self.modeltpl_cfg['n_knowledges']), peek_cur=False)
        z = (
            z.view(bs, self.modeltpl_cfg['n_knowledges'], seqlen, self.modeltpl_cfg['hidden_size'])  # unpack dimensions
            .transpose(1, 2)  # (bs, seqlen, n_know, d_model)
            .contiguous()
            .view(bs, seqlen, -1)
        )
        c_score = (
            c_score.view(bs, self.modeltpl_cfg['n_knowledges'], self.modeltpl_cfg['num_heads'], seqlen, seqlen)  # unpack dimensions
            .permute(0, 2, 3, 1, 4)  # (bs, n_heads, seqlen, n_know, seqlen)
            .contiguous()
        )
        if not is_train:
            return z, q_emb
        else:
            query = q_emb[:, n - 1 :, :]
            h = self.readout(z[:, : query.size(1), :], query)
            y = torch.sigmoid(self.out(torch.cat([query, h], dim=-1))).squeeze(-1)

            return y, z, q_emb, (exer_diff**2).mean() * 1e-3, (e_score, c_score)
        
    def readout(self, z, query):
        bs, seqlen, _ = query.size()
        key = (
            self.know_params[None, None, :, :]
            .expand(bs, seqlen, -1, -1)
            .view(bs * seqlen, self.modeltpl_cfg['n_knowledges'], -1)
        )
        value = z.reshape(bs * seqlen, self.modeltpl_cfg['n_knowledges'], -1)

        beta = torch.matmul(
            key,
            query.reshape(bs * seqlen, -1, 1),
        ).view(bs * seqlen, 1, self.modeltpl_cfg['n_knowledges'])
        alpha = torch.softmax(beta, -1)

        return torch.matmul(alpha, value).view(bs, seqlen, -1)
    
    def embedding(self, cpt_seq, label_seq, exer_seq=None):
        cpt_emb = self.cpt_embed(cpt_seq)
        label_emb = self.label_embed(label_seq.long()) + cpt_emb

        exer_diff = 0.0
        exer_diff = self.exer_diff_embed(exer_seq)

        cpt_diff_emb = self.cpt_diff_embed(cpt_seq)  # Concept variations
        cpt_emb += exer_diff * cpt_diff_emb

        label_diff_emb = self.label_diff_embed(label_seq.long()) + cpt_diff_emb
        label_emb += exer_diff * label_diff_emb

        return cpt_emb, label_emb, exer_diff

    @torch.no_grad()
    def predict(self, n=1, **kwargs):  
        z, q_emb, = self(**kwargs, is_train=False)
        query = q_emb[:, n-1 :, :]
        h = self.readout(z[:, : query.size(1), :], query)
        y_pd = torch.sigmoid(self.out(torch.cat([query, h], dim=-1))).squeeze(-1)

        y_pd = y_pd if n == 1 else y_pd[:, :-n+1]
        y_pd = y_pd[kwargs['mask_seq'][:, n-1:] == 1]

        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, n-1:]
            y_gt = y_gt[kwargs['mask_seq'][:, n-1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_loss(self, **kwargs):
        y_pd, _, _, reg_loss, _ = self(**kwargs, is_train=True)
        y_pd = y_pd[kwargs['mask_seq'] == 1]

        y_gt = kwargs['label_seq']
        y_gt = y_gt[kwargs['mask_seq'] == 1]

        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt, reduction="mean"
        ) + reg_loss
        return {
            'loss_main': loss
        }

    def get_cl_loss(self, **kwargs):   
        bs = kwargs['exer_seq'].size(0)
        lens = (kwargs['mask_seq'] > 0).sum(dim=1)
        minlen = lens.min().item()
        if minlen < MIN_SEQ_LEN:
            return self.get_loss(**kwargs)
        
        # Data Augmentation
        cpt_seq_ = kwargs['cpt_unfold_seq'].clone()
        label_seq_ = kwargs['label_seq'].clone()
        exer_seq_ = kwargs['exer_seq'].clone()
        # Manipulate order: Swap Adjacent Items
        for b in range(bs):
            idx = random.sample(range(lens[b] - 1), max(1, int(lens[b] * self.modeltpl_cfg['dropout'])))
            for i in idx:
                cpt_seq_[b, i], cpt_seq_[b, i + 1] = cpt_seq_[b, i + 1], cpt_seq_[b, i]
                label_seq_[b, i], label_seq_[b, i + 1] = label_seq_[b, i + 1], label_seq_[b, i]
                exer_seq_[b, i], exer_seq_[b, i + 1] = exer_seq_[b, i + 1], exer_seq_[b, i]
        # Hard negative 
        label_seq_flip = kwargs['label_seq'].clone() if self.modeltpl_cfg['hard_negative'] else label_seq_
        # Manipulate score: Flip Response
        for b in range(bs):
            idx = random.sample(range(lens[b] - 1), max(1, int(lens[b] * self.modeltpl_cfg['dropout'])))
            for i in idx:
                label_seq_flip[b, i] = 1 - label_seq_flip[b, i]
        if not self.modeltpl_cfg['hard_negative']: label_seq_ = label_seq_flip

        # Model loss
        logits, z_1, q_emb, reg_loss, _ = self(**kwargs, is_train=True)
        logits_masked = logits[kwargs['mask_seq'] == 1]

        _, z_2, *_ = self(exer_seq_, cpt_seq_, label_seq_, kwargs['mask_seq'], is_train=True)
        if self.modeltpl_cfg['hard_negative']:
            _, z_3, *_ = self(kwargs['exer_seq'], kwargs['cpt_unfold_seq'], label_seq_flip, kwargs['mask_seq'], is_train=True)

        # CL loss
        input = self.sim(z_1[:, :minlen, :], z_2[:, :minlen, :])
        if self.modeltpl_cfg['hard_negative']:
            hard_neg = self.sim(z_1[:, :minlen, :], z_3[:, :minlen, :])
            input = torch.cat([input, hard_neg], dim=1)
        target = (
            torch.arange(kwargs['label_seq'].size(0))[:, None]
            .to(self.know_params.device)
            .expand(-1, minlen)
        )
        cl_loss = F.cross_entropy(
            input=input, target=target
        )

        # Prediction loss
        labels_masked = kwargs['label_seq'][kwargs['mask_seq'] == 1]
        pred_loss = F.binary_cross_entropy(
            input=logits_masked, target=labels_masked, reduction="mean"
        )
        
        for i in range(1, self.modeltpl_cfg["prediction_window"]):
            y_gt = kwargs['label_seq'][:, i:]
            y_gt = y_gt[kwargs['mask_seq'][:, i:] == 1]

            query = q_emb[:, i:, :]
            h = self.readout(z_1[:, : query.size(1), :], query)
            y_pt = torch.sigmoid(self.out(torch.cat([query, h], dim=-1))).squeeze(-1)
            y_pd = y_pd[kwargs['mask_seq'][:, i:] == 1]

            pred_loss += F.binary_cross_entropy(
            input=y_pt, target=y_gt, reduction="mean"
        )
        pred_loss /= self.modeltpl_cfg["prediction_window"]

        return {
            'loss_main': pred_loss + cl_loss * self.modeltpl_cfg['lambda_cl'] + reg_loss,
        }

    def get_loss_dict(self, **kwargs):
        if self.modeltpl_cfg['cl_loss']:
            return self.get_cl_loss(**kwargs)
        else:
            return self.get_loss(**kwargs)

    def sim(self, z1, z2):
        bs, seqlen, _ = z1.size()
        z1 = z1.unsqueeze(1).view(bs, 1, seqlen, self.modeltpl_cfg['n_knowledges'], -1)
        z2 = z2.unsqueeze(0).view(1, bs, seqlen, self.modeltpl_cfg['n_knowledges'], -1)
        if self.modeltpl_cfg['projection_alhead_cl']:
            z1 = self.proj(z1)
            z2 = self.proj(z2)
        return F.cosine_similarity(z1.mean(-2), z2.mean(-2), dim=-1) / 0.05

class DTransformerLayer(nn.Module):
    def __init__(self, d_model, n_heads, dropout, kq_same=True):
        super().__init__()
        self.masked_attn_head = MultiHeadAttention(d_model, n_heads, kq_same)
        self.dropout_rate = dropout
        self.dropout = nn.Dropout(dropout)
        self.layer_norm = nn.LayerNorm(d_model)

    def device(self):
        return next(self.parameters()).device
    
    def forward(self, query, key, values, lens, peek_cur=False):
        seqlen = query.size(1)
        mask = torch.ones(seqlen, seqlen).tril(0 if peek_cur else -1)  
        mask = mask.bool()[None, None, :, :].to(self.device())

        # mask manipulation: Drop Item
        if self.training:
            mask = mask.expand(query.size(0), -1, -1, -1).contiguous() 

            for b in range(query.size(0)):
                # sample for each batch
                if lens[b] < MIN_SEQ_LEN:
                    # skip for short sequences
                    continue
                idx = random.sample(
                    range(lens[b] - 1), max(1, int(lens[b] * self.dropout_rate))
                )
                for i in idx:
                    mask[b, :, i + 1 :, i] = 0
        
        # apply transformer layer
        query_, scores = self.masked_attn_head(
            query, key, values, mask, maxout=not peek_cur
        )
        query = query + self.dropout(query_)

        return self.layer_norm(query), scores

class MultiHeadAttention(nn.Module):
    def __init__(self, d_model, n_heads, kq_same=True, bias=True):
        super().__init__()
        self.d_model = d_model
        self.d_k = d_model // n_heads
        self.h = n_heads

        self.q_linear = nn.Linear(d_model, d_model, bias=bias)
        if kq_same:
            self.k_linear = self.q_linear
        else:
            self.k_linear = nn.Linear(d_model, d_model, bias=bias)
        self.v_linear = nn.Linear(d_model, d_model, bias=bias)

        self.out_proj = nn.Linear(d_model, d_model, bias=bias)
        self.gammas = nn.Parameter(torch.zeros(n_heads, 1, 1))
        torch.nn.init.xavier_uniform_(self.gammas)
    
    def forward(self, q, k, v, mask, maxout=False):
        bs = q.size(0)
        # perform linear operation and split into h heads
        q = self.q_linear(q).view(bs, -1, self.h, self.d_k)
        k = self.k_linear(k).view(bs, -1, self.h, self.d_k)
        v = self.v_linear(v).view(bs, -1, self.h, self.d_k)

        # transpose to get dimensions bs * h * sl * d_k
        k, q, v = k.transpose(1, 2), q.transpose(1, 2), v.transpose(1, 2)

        # calculate attention using function we will define next
        v_, scores = attention(q, k, v, mask, self.gammas, maxout)

        # concatenate heads and put through final linear layer
        concat = v_.transpose(1, 2).contiguous().view(bs, -1, self.d_model)

        output = self.out_proj(concat)

        return output, scores

def attention(q, k, v, mask, gamma=None, maxout=False):
    # attention score with scaled dot production
    d_k = k.size(-1)
    scores = torch.matmul(q, k.transpose(-2, -1)) / math.sqrt(d_k)
    bs, head, seqlen, _ = scores.size()

    # include temporal effect
    if gamma is not None:
        x1 = torch.arange(seqlen).float().expand(seqlen, -1).to(gamma.device)
        x2 = x1.transpose(0, 1).contiguous()

        with torch.no_grad():
            scores_ = scores.masked_fill(mask == 0, -1e32)
            scores_ = F.softmax(scores_, dim=-1)

            distcum_scores = torch.cumsum(scores_, dim=-1)
            disttotal_scores = torch.sum(scores_, dim=-1, keepdim=True)
            position_effect = torch.abs(x1 - x2)[None, None, :, :]  
            dist_scores = torch.clamp(  
                (disttotal_scores - distcum_scores) * position_effect, min=0.0
            )
            dist_scores = dist_scores.sqrt().detach()

        gamma = -1.0 * gamma.abs().unsqueeze(0)
        total_effect = torch.clamp((dist_scores * gamma).exp(), min=1e-5, max=1e5)

        scores *= total_effect
    
    # normalize attention score
    scores.masked_fill_(mask == 0, -1e32)
    scores = F.softmax(scores, dim=-1)
    scores = scores.masked_fill(mask == 0, 0)  # set to hard zero to avoid leakage

    # max-out scores (bs, n_heads, seqlen, seqlen)
    if maxout:
        scale = torch.clamp(1.0 / scores.max(dim=-1, keepdim=True)[0], max=5.0)
        scores *= scale
    
    # calculate output
    output = torch.matmul(scores, v)
    return output, scores




import os

import torch

from ..common import EduDataTPL
import pandas as pd


class ECDDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_FilterRecords4CD', 'M2C_ReMapId', 'M2C_RandomDataSplit4CD',
                             'M2C_GenQMat'],
    }

    def process_load_data_from_middata(self):
        super().process_load_data_from_middata()
        self.final_kwargs['qqq_group_list'] = self.read_QQQ_group(self.cfg)
        qqq_list = self.df_stu['qqq_seq:token_seq'].to_list()
        self.QQQ_list = torch.tensor(qqq_list)
        self.final_kwargs['qqq_list'] = self.QQQ_list
        qqq_count = torch.max(self.QQQ_list) + 1
        self.cfg['datatpl_cfg']['dt_info']['qqq_count'] = qqq_count

    def get_extra_data(self):
        return {
            'qqq_group_list': self.final_kwargs['qqq_group_list'],
            'Q_mat': self.final_kwargs['Q_mat'],
            'qqq_list': self.final_kwargs['qqq_list']
        }

    def read_QQQ_group(self, cfg):
        group_path = f'{cfg.frame_cfg.data_folder_path}/middata/{cfg.dataset}_QQQ-group.csv'
        assert os.path.exists(group_path)
        df_QQQ_group = pd.read_csv(group_path, encoding='utf-8', usecols=['qqq_id:token', 'group_id:token'])
        gp = df_QQQ_group.groupby("group_id:token")['qqq_id:token']
        gps = gp.groups
        gps_list = []
        for k, v in gps.items():
            gps_list.append(list(v))
        return gps_list



from edustudio.utils.common import UnifyConfig
import logging
import os


class BaseRaw2Mid(object):
    def __init__(self, dt, rawpath, midpath) -> None:
        self.dt = dt
        self.rawpath = rawpath
        self.midpath = midpath
        self.logger = logging.getLogger("edustudio")
        if not os.path.exists(self.midpath):
            os.makedirs(self.midpath)

    @classmethod
    def from_cfg(cls, cfg: UnifyConfig):
        rawdata_folder_path = f"{cfg.frame_cfg.data_folder_path}/rawdata"
        middata_folder_path = f"{cfg.frame_cfg.data_folder_path}/middata"
        dt = cfg.dataset
        return cls(dt, rawdata_folder_path, middata_folder_path)
    
    def process(self, **kwargs):
        self.logger.info(f"{self.__class__.__name__} start !")
    
    @classmethod
    def from_cli(cls, dt, rawpath="./rawpath", midpath="./midpath"):
        obj = cls(dt, rawpath, midpath)
        obj.process()

# __cli_api_dict__ = {
#     'BaseRaw2Mid': BaseRaw2Mid.from_cli
# }



from .base_mid2cache import BaseMid2Cache
import pandas as pd
import numpy as np
from itertools import chain


class M2C_FilteringRecordsByAttr(BaseMid2Cache):
    """Commonly used by Fair Models, and Filtering Students without attribute values
    """
    default_cfg = {
        'filter_stu_attrs': ['gender:token']
    }

    def process(self, **kwargs):
        df_stu = kwargs['df_stu']
        df = kwargs['df']
        df_stu = df_stu[df_stu[self.m2c_cfg['filter_stu_attrs']].notna().all(axis=1)].reset_index(drop=True)
        df = df[df['stu_id:token'].isin(df_stu['stu_id:token'])].reset_index(drop=True)

        kwargs['df'] = df
        kwargs['df_stu'] = df_stu

        return kwargs
    





    


from edustudio.utils.common import UnifyConfig
from ..common.base_mid2cache import BaseMid2Cache
from edustudio.datatpl.utils import SpliterUtil
from sklearn.model_selection import StratifiedKFold, KFold
from itertools import chain
import numpy as np
import pandas as pd


class M2C_RandomDataSplit4CD(BaseMid2Cache):
    default_cfg = {
        'seed': 2023,
        "divide_scale_list": [7,1,2],
    }

    def __init__(self, m2c_cfg, n_folds, is_dataset_divided) -> None:
        super().__init__(m2c_cfg)
        self.n_folds = n_folds
        self.is_dataset_divided = is_dataset_divided

        assert is_dataset_divided is False

    def _check_params(self):
        super()._check_params()
        assert 2 <= len(self.m2c_cfg['divide_scale_list']) <= 3
        assert sum(self.m2c_cfg['divide_scale_list']) == 10

    @classmethod
    def from_cfg(cls, cfg: UnifyConfig):
        m2c_cfg = cfg.datatpl_cfg.get(cls.__name__)
        n_folds = cfg.datatpl_cfg.n_folds
        is_dataset_divided = cfg.datatpl_cfg.is_dataset_divided
        return cls(m2c_cfg, n_folds, is_dataset_divided)

    def process(self, **kwargs):
        df = kwargs['df']

        if self.n_folds == 1:
            assert kwargs.get("df_train", None) is None
            assert kwargs.get("df_valid", None) is None
            assert kwargs.get("df_test", None) is None
            df_train, df_valid, df_test = self.one_fold_split(df)
            kwargs['df_train_folds'] = [df_train]
            kwargs['df_valid_folds'] = [df_valid] if df_valid is not None else []
            kwargs['df_test_folds'] = [df_test]
        else:
            df_train_list, df_test_list = self.multi_fold_split(df)
            kwargs['df_train_folds'] = df_train_list
            kwargs['df_test_folds'] = df_test_list

        return kwargs

    def one_fold_split(self, df):
        return SpliterUtil.divide_data_df_one_fold(
            df, divide_scale_list=self.m2c_cfg['divide_scale_list'], seed=self.m2c_cfg['seed']
        )

    def multi_fold_split(self, df):
        skf = KFold(n_splits=int(self.n_folds), shuffle=True, random_state=self.m2c_cfg['seed'])
        splits = skf.split(df)

        train_list, test_list = [], []
        for train_index, test_index in splits:
            train_df = df.iloc[train_index].reset_index(drop=True)
            test_df = df.iloc[test_index].reset_index(drop=True)
            train_list.append(train_df)
            test_list.append(test_df)
        return train_list, test_list

    def set_dt_info(self, dt_info, **kwargs):
        if 'feats_group' not in kwargs:
            if 'stu_id:token' in kwargs['df'].columns:
                dt_info['stu_count'] = int(kwargs['df']['stu_id:token'].max() + 1)
            if 'exer_id:token' in kwargs['df'].columns:
                dt_info['exer_count'] = int(kwargs['df']['exer_id:token'].max() + 1)
            if kwargs.get('df_exer', None) is not None:
                if 'cpt_seq:token_seq' in kwargs['df_exer']:
                    dt_info['cpt_count'] = np.max(list(chain(*kwargs['df_exer']['cpt_seq:token_seq'].to_list()))) + 1
        else:
            feats_group = kwargs['feats_group']
            stu_id_group, exer_id_group, cpt_id_group = [], [], []
            for gp in feats_group:
                if 'stu_id:token' in gp:
                    stu_id_group = set(gp)
                if 'exer_id:token' in gp:
                    exer_id_group = set(gp)
                if 'cpt_seq:token_seq' in gp:
                    cpt_id_group = set(gp)
            for df in kwargs.values():
                if type(df) is pd.DataFrame:
                    for col in df.columns:
                        if col in stu_id_group:
                            if col.split(":")[-1] == 'token':
                                dt_info['stu_count'] = max(dt_info.get('stu_count', -1), df[col].max() + 1)
                            else:
                                dt_info['stu_count'] = max(dt_info.get('stu_count', -1), np.max(list(chain(*df[col].to_list()))) + 1)
                        if col in exer_id_group:
                            if col.split(":")[-1] == 'token':
                                dt_info['exer_count'] = max(dt_info.get('exer_count', -1), df[col].max() + 1)
                            else:
                                dt_info['exer_count'] = max(dt_info.get('exer_count', -1), np.max(list(chain(*df[col].to_list()))) + 1)
                        if col in cpt_id_group:
                            if col.split(":")[-1] == 'token':
                                dt_info['cpt_count'] = max(dt_info.get('cpt_count', -1), df[col].max() + 1)
                            else:
                                dt_info['cpt_count'] = max(dt_info.get('cpt_count', -1), np.max(list(chain(*df[col].to_list()))) + 1)



import torch
import torch.nn as nn
from typing import Union
import torch.nn.functional as F

class PosLinear(nn.Linear):
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        weight = 2 * F.relu(1 * torch.neg(self.weight)) + self.weight
        return F.linear(input, weight, self.bias)

class Identity(nn.Module):
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, inputs):
        return inputs

class ActivationUtil(object):
    @staticmethod
    def get_common_activation_layer(act_obj: Union[str, nn.Module] = "relu") -> nn.Module:
        if isinstance(act_obj, str):
            if act_obj.lower() == 'relu':
                return nn.ReLU(inplace=True)
            elif act_obj.lower() == 'sigmoid':
                return nn.Sigmoid()
            elif act_obj.lower() == 'linear':
                return Identity()
            elif act_obj.lower() == 'prelu':
                return nn.PReLU()
            elif act_obj.lower() == 'elu':
                return nn.ELU(inplace=True)
            elif act_obj.lower() == 'leakyrelu':
                return nn.LeakyReLU(0.2, inplace=True)
        else:
            return act_obj()


class PosMLP(nn.Module):
    """
        The Multi Layer Perceptron (MLP)
        note: output layer has no activation function, output layer has batch norm and dropout
    """

    def __init__(self, input_dim: int, output_dim: int, dnn_units: Union[list, tuple],
                 activation: Union[str, nn.Module, list] = 'relu', dropout_rate: float = 0.0,
                 use_bn: bool = False, device='cpu'):
        super().__init__()
        self.use_bn = use_bn
        dims_list = [input_dim] + list(dnn_units) + [output_dim]
        if type(activation) is list:
            assert len(activation) == len(dnn_units)

        self.linear_units_list = nn.ModuleList(
            [PosLinear(dims_list[i], dims_list[i + 1], bias=True) for i in range(len(dims_list) - 1)]
        )
        self.act_units_list = nn.ModuleList(
            [ActivationUtil.get_common_activation_layer(activation)] * len(dnn_units)
            if type(activation) is not list else [ActivationUtil.get_common_activation_layer(i) for i in activation]
        )
        self.dropout_layer = nn.Dropout(dropout_rate)

        if use_bn is True:
            self.bn_units_list = nn.ModuleList(
                [nn.BatchNorm1d(dims_list[i + 1]) for i in range(len(dims_list) - 1)]
            )
            assert len(self.linear_units_list) == len(self.bn_units_list)
        assert len(self.linear_units_list) == len(self.act_units_list) + 1
        self.to(device)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        tmp = input
        for i in range(len(self.act_units_list)):
            tmp = self.linear_units_list[i](tmp)
            if self.use_bn is True:
                tmp = self.bn_units_list[i](tmp)
            tmp = self.act_units_list[i](tmp)
            tmp = self.dropout_layer(tmp)
        tmp = self.linear_units_list[-1](tmp)
        if self.use_bn is True:
            tmp = self.bn_units_list[-1](tmp)
        output = tmp
        return output




class MLP(nn.Module):
    """
        The Multi Layer Perceptron (MLP)
        note: output layer has no activation function, output layer has batch norm and dropout
    """

    def __init__(self, input_dim: int, output_dim: int, dnn_units: Union[list, tuple],
                 activation: Union[str, nn.Module, list] = 'relu', dropout_rate: float = 0.0,
                 use_bn: bool = False, device='cpu'):
        super().__init__()
        self.use_bn = use_bn
        dims_list = [input_dim] + list(dnn_units) + [output_dim]
        if type(activation) is list:
            assert len(activation) == len(dnn_units)

        self.linear_units_list = nn.ModuleList(
            [nn.Linear(dims_list[i], dims_list[i + 1], bias=True) for i in range(len(dims_list) - 1)]
        )
        self.act_units_list = nn.ModuleList(
            [ActivationUtil.get_common_activation_layer(activation)] * len(dnn_units)
            if type(activation) is not list else [ActivationUtil.get_common_activation_layer(i) for i in activation]
        )
        self.dropout_layer = nn.Dropout(dropout_rate)

        if use_bn is True:
            self.bn_units_list = nn.ModuleList(
                [nn.BatchNorm1d(dims_list[i + 1]) for i in range(len(dims_list) - 1)]
            )
            assert len(self.linear_units_list) == len(self.bn_units_list)
        assert len(self.linear_units_list) == len(self.act_units_list) + 1
        self.to(device)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        tmp = input
        for i in range(len(self.act_units_list)):
            tmp = self.linear_units_list[i](tmp)
            if self.use_bn is True:
                tmp = self.bn_units_list[i](tmp)
            tmp = self.act_units_list[i](tmp)
            tmp = self.dropout_layer(tmp)
        tmp = self.linear_units_list[-1](tmp)
        if self.use_bn is True:
            tmp = self.bn_units_list[-1](tmp)
        output = tmp
        return output


from .base_traintpl import BaseTrainTPL
from edustudio.utils.common import UnifyConfig, set_same_seeds
import torch
from torch.utils.data import DataLoader
from edustudio.utils.callback import History
from collections import defaultdict
import numpy as np


class GDTrainTPL(BaseTrainTPL):
    default_cfg = {
        'device': 'cuda:0',
        'epoch_num': 100,
        'batch_size': 2048,
        'eval_batch_size': 2048,
        'num_workers': 0,

        'lr': 0.001,
        'optim': 'adam',
        'eps': 1e-8,
        'weight_decay': 0.0, 
    }

    def __init__(self, cfg: UnifyConfig):
        super().__init__(cfg)
        self.train_loader_list = []
        self.valid_loader_list = []
        self.test_loader_list = []
    
    def _get_optim(self, optimizer='adam', lr=0.001, weight_decay=0.0, eps=1e-8):
        """Get optimizer
        """
        if optimizer == "sgd":
            optim = torch.optim.SGD(self.model.parameters(), lr=lr, weight_decay=weight_decay, eps=eps)
        elif optimizer == "adam":
            optim = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay=weight_decay, eps=eps)
        elif optimizer == "adagrad":
            optim = torch.optim.Adagrad(self.model.parameters(), lr=lr, weight_decay=weight_decay, eps=eps)
        elif optimizer == "rmsprop":
            optim = torch.optim.RMSprop(self.model.parameters(), lr=lr, weight_decay=weight_decay, eps=eps)
        else:
            raise ValueError("unsupported optimizer")

        return optim

    def start(self):
        """entrypoint of starting a training process
        """
        super().start()
        self.build_loaders()

        best_metric_value_dict = defaultdict(list)

        for fold_id in range(self.datatpl_cfg['n_folds']):
            self.train_loader = self.train_loader_list[fold_id]
            self.train_loader.dataset.set_info_for_fold(fold_id)
            self.datatpl.set_info_for_fold(fold_id)

            self.valid_loader = None
            if self.datatpl.hasValidDataset:
                self.valid_loader = self.valid_loader_list[fold_id]
                self.valid_loader.dataset.set_info_for_fold(fold_id)
            self.test_loader = self.test_loader_list[fold_id]
            self.test_loader.dataset.set_info_for_fold(fold_id)

            extra_data = self.datatpl.get_extra_data() 

            set_same_seeds(self.traintpl_cfg['seed'])
            self.model = self.get_model_obj()
            self.model.build_cfg()
            self.model.add_extra_data(**extra_data)
            for evaltpl in self.evaltpls:
                evaltpl.add_extra_data(**extra_data)
            self.model.build_model()
            self.model._init_params()
            self.model.to(self.model.device)


            metrics = self.one_fold_start(fold_id)
            for m in metrics:
                best_metric_value_dict[m].append(metrics[m])
            self.logger.info("=="*10)
        
        best_metric_value_dict_mean = {}
        for metric_name, val_list in best_metric_value_dict.items():
            best_metric_value_dict_mean[metric_name] = np.mean(val_list)
            self.logger.info(f"All Fold Mean {metric_name} = {best_metric_value_dict_mean[metric_name]}")

        History.dump_json(best_metric_value_dict, f"{self.frame_cfg.temp_folder_path}/result-list.json")
        History.dump_json(best_metric_value_dict_mean, f"{self.frame_cfg.temp_folder_path}/result.json")
    
    def one_fold_start(self, fold_id):
        """training process of one one fold

        Args:
            fold_id (int): fold id
        """
        self.logger.info(f"====== [FOLD ID]: {fold_id} ======")

    def batch_dict2device(self, batch_dict):
        dic = {}
        for k, v in batch_dict.items():
            if not isinstance(v, list):
                dic[k] = v.to(self.traintpl_cfg['device'])
            else:
                dic[k] = v
        return dic

    def build_loaders(self):
        """build dataloaders
        """
        batch_size = self.traintpl_cfg['batch_size']
        num_workers = self.traintpl_cfg['num_workers']
        eval_batch_size = self.traintpl_cfg['eval_batch_size']
        if hasattr(self.datatpl, 'build_dataloaders'):
            self.train_loader_list, self.valid_loader_list, self.test_loader_list = self.datatpl.build_dataloaders()
        else:
            train_dt_list, valid_dt_list, test_dt_list = self.datatpl.build_datasets()
            for fid in range(self.datatpl_cfg['n_folds']):
                train_loader = DataLoader(dataset=train_dt_list[fid], shuffle=True, batch_size=batch_size, 
                                          num_workers=num_workers, collate_fn=train_dt_list[fid].collate_fn)
                self.train_loader_list.append(train_loader)
                if self.hasValidDataset:
                    valid_loader = DataLoader(dataset=valid_dt_list[fid], shuffle=False, batch_size=eval_batch_size,
                                               num_workers=num_workers, collate_fn=valid_dt_list[fid].collate_fn)
                    self.valid_loader_list.append(valid_loader)
                test_loader = DataLoader(dataset=test_dt_list[fid], shuffle=False, batch_size=eval_batch_size, 
                                         num_workers=num_workers, collate_fn=test_dt_list[fid].collate_fn)
                self.test_loader_list.append(test_loader)



import importlib
from .base_datatpl import BaseDataTPL
from edustudio.utils.common import UnifyConfig


class BaseProxyDataTPL(object):
    """The basic protocol for implementing a proxy data template
    """
    default_cfg = {'backbone_datatpl_cls': 'BaseDataTPL'}

    @classmethod
    def from_cfg_proxy(cls, cfg):
        """an interface to instantiate a proxy model

        Args:
            cfg (UnifyConfig): the global config object

        Returns:
           BaseProxyDataTPL
        """
        backbone_datatpl_cls = cls.get_backbone_cls(cfg.datatpl_cfg.backbone_datatpl_cls)
        new_cls = cls.get_new_cls(p_cls=backbone_datatpl_cls)
        return new_cls.from_cfg(cfg)

    @classmethod
    def get_backbone_cls(cls, backbone_datatpl_cls):
        if isinstance(backbone_datatpl_cls, str):
            backbone_datatpl_cls = importlib.import_module('edustudio.datatpl').\
                __getattribute__(backbone_datatpl_cls)
        elif issubclass(backbone_datatpl_cls, BaseDataTPL):
            backbone_datatpl_cls = backbone_datatpl_cls
        else:
            raise ValueError(f"Unknown type of backbone_datatpl_cls: {backbone_datatpl_cls}")
        return backbone_datatpl_cls
    
    @classmethod
    def get_new_cls(cls, p_cls):
        """dynamic inheritance

        Args:
            p_cls (BaseModel): parent class

        Returns:
            BaseProxyModel: A inherited class
        """
        new_cls = type(cls.__name__ + "_proxy", (cls, p_cls), {})
        return new_cls

    @classmethod
    def get_default_cfg(cls, backbone_datatpl_cls, **kwargs):
        """Get the final default_cfg

        Args:
            backbone_datatpl_cls (BaseDataTPL): backbone data template class name

        Returns:
            UnifyConfig: the final default config object
        """
        bb_cls = None
        if backbone_datatpl_cls is not None:
            bb_cls = cls.get_backbone_cls(backbone_datatpl_cls)
        else:
            for _cls in cls.__mro__:
                if not hasattr(_cls, 'default_cfg'):
                    break
                bb_cls = _cls.default_cfg.get('backbone_datatpl_cls', None)
                if bb_cls is not None: break
            assert bb_cls is not None
            bb_cls = cls.get_backbone_cls(bb_cls)
        
        cfg = UnifyConfig()
        cfg.backbone_datatpl_cls = bb_cls
        cfg.backbone_datatpl_cls_name = bb_cls.__name__
        new_cls = cls.get_new_cls(p_cls=bb_cls)
        for _cls in new_cls.__mro__:
            if not hasattr(_cls, 'default_cfg'):
                break
            if issubclass(_cls, BaseProxyDataTPL):
                cfg.update(_cls.default_cfg, update_unknown_key_only=True)
            else:
                cfg.update(_cls.get_default_cfg(**kwargs), update_unknown_key_only=True)
                break
        return cfg



from .base_mid2cache import BaseMid2Cache
import numpy as np
import pandas as pd
from itertools import chain
import torch
from edustudio.utils.common import set_same_seeds, tensor2npy
from tqdm import tqdm

class M2C_FillMissingQ(BaseMid2Cache):
    default_cfg = {
        'Q_fill_type': "None",
        'params_topk': 5, 
        'params_votek': 2,
    }

    def __init__(self, m2c_cfg, cfg) -> None:
        self.logger = cfg.logger
        self.m2c_cfg = m2c_cfg
        self.cfg = cfg

    @classmethod
    def from_cfg(cls, cfg):
        return cls(cfg.datatpl_cfg.get(cls.__name__), cfg)

    def process(self, **kwargs):
        dt_info = kwargs['dt_info']
        self.user_count = dt_info['stu_count']
        self.item_count = dt_info['exer_count']
        self.cpt_count = dt_info['cpt_count']
        self.df_Q = kwargs['df_exer'][['exer_id:token', 'cpt_seq:token_seq']]

        Q_mat = kwargs['Q_mat']
        missing_Q_mat = kwargs['missing_Q_mat']

        self.filling_Q_mat_list = []
        for df_train in kwargs['df_train_folds']:
            if (missing_Q_mat.sum(dim=1) == 0).sum() > 0:
                if self.m2c_cfg['Q_fill_type'] == "sim_dist_for_by_exer":
                    fill_df_Q = self.fill_df_Q_by_sim_dist(
                        df_train, kwargs['missing_df_Q'], 
                        params_topk=self.m2c_cfg['params_topk'], 
                        params_votek=self.m2c_cfg['params_votek']
                    )
                    fill_Q_mat = self.get_Q_mat_from_df_arr(fill_df_Q, self.item_count, self.cpt_count)
                    self.filling_Q_mat_list.append(fill_Q_mat)
                elif self.m2c_cfg['Q_fill_type'] == "None":
                    self.filling_Q_mat_list.append(missing_Q_mat)
                else:
                    raise ValueError(f"unknown Q_fill_type: {self.m2c_cfg['Q_fill_type']}")
            else:
                self.filling_Q_mat_list.append(Q_mat)

        kwargs['filling_Q_mat_list'] = self.filling_Q_mat_list
        return kwargs

    def get_Q_mat_from_df_arr(self, df_Q_arr, item_count, cpt_count):
        Q_mat = np.zeros((item_count, cpt_count), dtype=np.int64)
        for _, item in df_Q_arr.iterrows(): Q_mat[item['exer_id:token'], item['cpt_seq:token_seq']] = 1
        return Q_mat

    def fill_df_Q_by_sim_dist(self, df_interaction, df_Q_left, params_topk=5, params_votek=2):
        preserved_exers = df_Q_left['exer_id:token'].to_numpy()
        interact_mat = torch.zeros((self.user_count, self.item_count), dtype=torch.int8).to(self.cfg.traintpl_cfg['device'])
        idx = df_interaction[df_interaction['label:float'] == 1][['stu_id:token','exer_id:token']].to_numpy()
        interact_mat[idx[:,0], idx[:,1]] = 1  
        idx = df_interaction[df_interaction['label:float'] != 1][['stu_id:token','exer_id:token']].to_numpy()
        interact_mat[idx[:,0], idx[:,1]] = -1 

        interact_mat = interact_mat.T

        sim_mat = torch.zeros((self.item_count, self.item_count))
        missing_iids = np.array(list(set(np.arange(self.item_count)) - set(preserved_exers)))
        for iid in tqdm(missing_iids, desc="[FILL_Q_MAT] compute sim_mat", ncols=self.cfg.frame_cfg['TQDM_NCOLS']):
            temp = interact_mat[iid] != 0
            same_mat =  interact_mat[iid] == interact_mat
            bool_mat = (temp) & (interact_mat != 0) 
            same_mat[~bool_mat] = False
            sim_mat[iid] = same_mat.sum(dim=1) / (temp).sum()
            sim_mat[iid, bool_mat.sum(dim=1) == 0] = 0.0
            sim_mat[iid, iid] = -1.0
            sim_mat[iid, missing_iids] = -1.0

        assert torch.isnan(sim_mat).sum() == 0

        _, topk_mat_idx = torch.topk(sim_mat, dim=1, k=params_topk, largest=True, sorted=True)
        topk_mat_idx = tensor2npy(topk_mat_idx)

        index_df_Q = df_Q_left.set_index('exer_id:token')
        missing_iid_fill_cpts = {}
        for iid in tqdm(missing_iids, desc="[FILL_Q_MAT] fill process", ncols=self.cfg.frame_cfg['TQDM_NCOLS']):
            count_dict = dict(zip(*np.unique(
                list(chain(*[index_df_Q.loc[iid2]['cpt_seq:token_seq'] for iid2 in topk_mat_idx[iid] if iid2 in preserved_exers])),
                return_counts=True,
            )))
            count_dict = sorted(count_dict.items(), key=lambda x: x[1], reverse=True)
            missing_iid_fill_cpts[iid] = [i[0] for i in count_dict[0:params_votek]]

        missing_fill_df_Q = pd.DataFrame(
            {'exer_id:token': list(missing_iid_fill_cpts.keys()),'cpt_seq:token_seq':list(missing_iid_fill_cpts.values())}
        )
        final_df_Q = pd.concat([df_Q_left, missing_fill_df_Q], axis=0, ignore_index=True)

        hit_ratio = 0
        t_Q = self.df_Q.set_index('exer_id:token')
        for iid in missing_iid_fill_cpts:
            if len(set(t_Q.loc[iid]['cpt_seq:token_seq']) & set(missing_iid_fill_cpts[iid])) > 0:
                hit_ratio += 1
        hit_ratio = hit_ratio / len(missing_iid_fill_cpts)

        self.logger.info(f"[FILL_Q] Hit_ratio={hit_ratio}")

        return final_df_Q



from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F


class CKT(GDBaseModel):
    default_cfg = {
        'hidden_size': 100,
        'drop_rate': 0,
        'k1': 6,
        'k2': 6
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']

        self.hidden_size = self.modeltpl_cfg['hidden_size']
        self.num_skills = self.n_item

    def build_model(self):
        skill_w = torch.empty(self.num_skills, self.hidden_size, device=self.device)
        torch.nn.init.xavier_uniform_(skill_w)
        self.skill_w = torch.nn.Parameter(skill_w)

        zeros = torch.zeros((self.num_skills, self.hidden_size), dtype=torch.float32, device=self.device)
        t1 = torch.cat([self.skill_w, zeros], dim=-1)
        t2 = torch.cat([zeros, self.skill_w], dim=-1)
        self.input_w = torch.cat([t1, t2], dim=0)

        self.cnn_block = CNNBlock(self.modeltpl_cfg['hidden_size'],
                                  self.modeltpl_cfg['k1'],
                                  self.modeltpl_cfg['k2'],
                                  self.modeltpl_cfg['drop_rate'],
                                  self.device).to(self.device)

        self.cnn = CNN(self.modeltpl_cfg['hidden_size'],
                       self.cnn_block,
                       self.device
                       ).to(self.device)

    def forward(self, exer_seq, label_seq, **kwargs):
        input_data, input_skill, l, next_id = self.data_helper(exer_seq, label_seq)
        skills = torch.nn.functional.embedding(input_skill.long(), self.skill_w)
        next_skill = torch.nn.functional.embedding(next_id.long(), self.skill_w)
        input_data = torch.nn.functional.embedding(input_data.long(), self.input_w)

        outputs = self.cnn(input_data, skills, l, next_skill)
        logits = torch.sum(outputs, dim=-1)
        y_pd = logits.sigmoid()
        return y_pd

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1][kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1][kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)

    def data_helper(self, exer_seq, label_seq):
        num_steps = self.datatpl_cfg['dt_info']['real_window_size']
        batch_size = len(exer_seq)
        input_data = torch.zeros((batch_size, num_steps), device=self.device)
        input_skill = torch.zeros((batch_size, num_steps), device=self.device)
        next_id = torch.zeros((batch_size, num_steps), device=self.device)
        l = torch.ones((batch_size, num_steps, self.num_skills), device=self.device)
        for i in range(batch_size):

            problem_ids = exer_seq[i]

            correctness = label_seq[i]
            correct_num = torch.zeros(self.num_skills, device=self.device)
            answer_count = torch.ones(self.num_skills, device=self.device)
            for j in range(len(problem_ids) - 1):
                problem_id = int(problem_ids[j])

                if (int(correctness[j]) == 0):
                    input_data[i, j] = problem_id + self.num_skills
                else:
                    input_data[i, j] = problem_id
                    correct_num[problem_id] += 1
                l[i, j] = correct_num / answer_count
                answer_count[problem_id] += 1
                input_skill[i, j] = problem_id
                next_id[i, j] = int(problem_ids[j + 1])
        return input_data, input_skill, l, next_id


def GLU(inputs, dim, device):
    sigmoid = nn.Sigmoid().to(device)
    dense1 = nn.Linear(inputs.size(-1), dim).to(device)
    dense2 = nn.Linear(inputs.size(-1), dim).to(device)

    r = sigmoid(dense1(inputs))

    output = dense2(inputs)

    output = output * r
    return output


class CNNBlock(nn.Module):
    def __init__(self, filter1, k1, k2, drop_rate, device):
        super(CNNBlock, self).__init__()

        self.filter1 = filter1
        self.k1 = k1
        self.k2 = k2
        self.drop_rate = drop_rate
        self.device = device

        self.w1 = nn.Parameter(torch.Tensor(k1, filter1, filter1))
        nn.init.xavier_uniform_(self.w1)

        self.w2 = torch.zeros(k2, filter1, filter1)

        self.res_w = nn.Parameter(torch.cat([self.w1, self.w2], dim=0))

        self.b = nn.Parameter(torch.Tensor(filter1))
        nn.init.uniform_(self.b)


    def forward(self, x):
        o1 = x
        o2 = F.conv1d(o1.permute([0, 2, 1]), self.res_w.permute([2, 1, 0]), stride=1, padding='same').permute(
            [0, 2, 1]) + self.b
        o2 = F.dropout(o2, p=self.drop_rate)
        o2 = GLU(o2, self.filter1, self.device)
        return o2 + x


class CNN(nn.Module):
    def __init__(self, hidden_size, cnn_block, device):
        super(CNN, self).__init__()

        self.device = device
        self.hidden_size = hidden_size

        self.cnn_block = cnn_block

    def forward(self, x, skills, individual, next_skill):
        att = self.cal_att(skills)

        x2 = torch.matmul(att, x)

        x = torch.cat([x, x2, individual], axis=-1)
        x = GLU(x, self.hidden_size, self.device)
        # x = self.glu(x)
        for i in range(1, 4):
            x = self.cnn_block(x)
        x = x * next_skill
        return x

    def cal_att(self, inputs):
        xnorm = torch.sqrt(torch.sum(torch.square(inputs), axis=2))

        xnorm1 = (xnorm.unsqueeze(1)).repeat([1, inputs.shape[1], 1])
        xnorm2 = (xnorm.unsqueeze(-1)).repeat([1, 1, inputs.shape[1]])

        x_x = inputs.matmul(inputs.permute([0, 2, 1]))

        outputs = torch.div(x_x, xnorm1 * xnorm2)
        diag_vals = torch.ones_like(outputs[0, :, :], device=self.device)
        tril = torch.tril(diag_vals, diagonal=2)
        sel = torch.eye(outputs.shape[1], device=self.device)
        sel2 = torch.eye(1, m=outputs.shape[1], device=self.device)
        sel3 = torch.zeros([outputs.shape[1] - 1, outputs.shape[1]], device=self.device)
        sel4 = torch.cat([sel2, sel3], dim=0)
        sel = sel - sel4
        tril = tril - sel

        masks = tril.unsqueeze(0).repeat([outputs.shape[0], 1, 1])
        paddings = torch.ones_like(masks, device=self.device) * (-2 ** 32 + 1)
        outputs = torch.where(masks == 0, paddings, outputs)

        outputs = F.softmax(outputs, dim=1)

        return outputs



r"""
IRT
##########################################

Reference Code:
    https://github.com/bigdata-ustc/EduCDM/tree/main/EduCDM/IRT

"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from ..gd_basemodel import GDBaseModel


class IRT(GDBaseModel):
    r"""
    IRT
    """
    default_cfg = {
        "a_range": -1.0, # disc range
        "diff_range": -1.0, # diff range
        "fix_a": False,
        "fix_c": True,
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        if self.modeltpl_cfg['a_range'] is not None and self.modeltpl_cfg['a_range']  < 0: self.modeltpl_cfg['a_range'] = None
        if self.modeltpl_cfg['diff_range'] is not None and self.modeltpl_cfg['diff_range'] < 0: self.modeltpl_cfg['diff_range'] = None

        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']

        # 确保c固定时，a一定不能固定
        if self.modeltpl_cfg['fix_c'] is False: assert self.modeltpl_cfg['fix_a'] is False

    def build_model(self):
        self.theta = nn.Embedding(self.n_user, 1) # student ability
        self.a = 0.0 if self.modeltpl_cfg['fix_a'] else nn.Embedding(self.n_item, 1) # exer discrimination
        self.b = nn.Embedding(self.n_item, 1) # exer difficulty
        self.c = 0.0 if self.modeltpl_cfg['fix_c'] else nn.Embedding(self.n_item, 1)

    def forward(self, stu_id, exer_id, **kwargs):
        theta = self.theta(stu_id)
        a = self.a(exer_id)
        b = self.b(exer_id)
        c = self.c if self.modeltpl_cfg['fix_c'] else self.c(exer_id).sigmoid()

        if self.modeltpl_cfg['diff_range'] is not None:
            b = self.modeltpl_cfg['diff_range'] * (torch.sigmoid(b) - 0.5)
        if self.modeltpl_cfg['a_range'] is not None:
            a = self.modeltpl_cfg['a_range'] * torch.sigmoid(a)
        else:
            a = F.softplus(a) # 让区分度大于0，保持单调性假设
        if torch.max(theta != theta) or torch.max(a != a) or torch.max(b != b):  # pragma: no cover
            raise ValueError('ValueError:theta,a,b may contains nan!  The diff_range or a_range is too large.')
        return self.irf(theta, a, b, c)

    @staticmethod
    def irf(theta, a, b, c, D=1.702):
        return c + (1 - c) / (1 + torch.exp(-D * a * (theta - b)))

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }




# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

import torch.nn as nn
from ..modeState import ModeState
import logging


class Callback(object):
    def __init__(self):
        self.model = None
        self.mode_state = None
        self.logger = None
        self.callback_list = None

    def set_model(self, model: nn.Module):
        self.model = model

    def set_state(self, mode_state: ModeState):
        self.mode_state = mode_state

    def set_logger(self, logger: logging.Logger):
        self.logger = logger

    def set_callback_list(self, callback_list):
        self.callback_list = callback_list

    def on_train_begin(self, logs=None, **kwargs):
        self.logger.info(f"[CALLBACK]-{self.__class__.__name__} has been registered!")

    def on_train_end(self, logs=None, **kwargs):
        pass

    def on_epoch_begin(self, epoch, logs=None, **kwargs):
        pass

    def on_epoch_end(self, epoch, logs=None, **kwargs):
        pass

    def on_train_batch_begin(self, batch, logs=None, **kwargs):
        pass

    def on_train_batch_end(self, batch, logs=None, **kwargs):
        pass



from .raw2mid import BaseRaw2Mid
import pandas as pd
r"""
R2M_ASSIST_2017
#####################################
ASSIST_2017 dataset preprocess
"""


class R2M_ASSIST_17(BaseRaw2Mid):
    """R2M_ASSIST_17 is to preprocess ASSISTment 2017 dataset"""
    def process(self):
        super().process()
        # pd.set_option("mode.chained_assignment", None)  # ingore warning


        # 读取数据 并显示
        df = pd.read_csv(f"{self.rawpath}/anonymized_full_release_competition_dataset.csv",encoding='utf-8',low_memory=False)

        # 进行映射
        df = df[['studentId','problemId','correct','startTime','timeTaken','MiddleSchoolId','InferredGender','skill','assignmentId']]

        # 处理skill，映射为skill_id
        knowledge_points = df['skill'].unique()
        knowledge_point_ids = {kp: idx for idx, kp in enumerate(knowledge_points, start=0)}
        df['skill_id'] = df['skill'].map(knowledge_point_ids)
        del df['skill']

        # 性别映射，空值保留
        gender_mapping = {'Male': 1, 'Female': 0}
        df['gender:float'] = df['InferredGender'].map(gender_mapping)
        del df['InferredGender']

        # 处理其他列
        def sort(data, column):
            '''将原始数据对指定列进行排序，并完成0-num-1映射'''
            if(column != 'startTime'):
                data .sort_values(column, inplace=True)
                value_mapping = {}
                new_value = 0
                for value in data[column].unique():
                    value_mapping[value] = new_value
                    new_value += 1
                new_column = f'new_{column}'
                data[new_column] = data[column].map(value_mapping)
                del data[column]
                
            # 单独处理 startTime，用字典映射
            else: 
                data = data.sort_values(by=['new_studentId', 'startTime'])

                user_mapping = {}
                user_count = {}

                def generate_mapping(row):
                    '''生成作答记录时间编号映射的函数'''
                    user_id = row['new_studentId']
                    timestamp = row['startTime']
                    if user_id not in user_mapping:
                        user_mapping[user_id] = {}
                        user_count[user_id] = 0
                    if timestamp not in user_mapping[user_id]:
                        user_mapping[user_id][timestamp] = user_count[user_id]
                        user_count[user_id] += 1
                    return user_mapping[user_id][timestamp]
                data['new_order_id'] = data.apply(generate_mapping, axis=1)
            return data
        df = sort(df,'studentId')
        df = sort(df,'assignmentId')
        df = sort(df,'problemId')
        df = sort(df,'MiddleSchoolId')
        df = sort(df,'startTime')
        
        # 修改列名及顺序
        df = df.rename(columns = {'new_studentId' : 'stu_id:token', 'new_assignmentId':'assignment_id:token_seq','new_problemId' : 'exer_id:token',
                                  'correct' : 'label:float','new_MiddleSchoolId':'school_id:token','new_order_id':'order_id:float',
                                  'startTime':'start_timestamp:float', 'timeTaken':'cost_time:float','skill_id':'cpt_seq:token_seq'})
        new_column_order = ['stu_id:token','exer_id:token','label:float','start_timestamp:float','cost_time:float','order_id:float',
                            'school_id:token', 'gender:float','cpt_seq:token_seq','assignment_id:token_seq']
        df = df.reindex(columns=new_column_order)

        # df_inter 的相关处理
        df_inter = df[['stu_id:token','exer_id:token','label:float','start_timestamp:float','cost_time:float','order_id:float']]
        df_inter.drop_duplicates(inplace=True)
        df_inter .sort_values('stu_id:token', inplace=True)
        
        # df_user 相关处理
        df_user = df[['stu_id:token','school_id:token','gender:float']]
        df_user.drop_duplicates(inplace=True)
        df_user .sort_values('stu_id:token', inplace=True)

        # df_exer 相关处理

        # 处理列名
        df_exer = df[['exer_id:token','cpt_seq:token_seq','assignment_id:token_seq']]
        df_exer.sort_values(by='exer_id:token', inplace=True)
        df_exer.drop_duplicates(inplace=True)

        # 合并 cpt_seq
        grouped_skills = df_exer[['exer_id:token','cpt_seq:token_seq']]
        grouped_skills.drop_duplicates(inplace=True)
        grouped_skills.sort_values(by='cpt_seq:token_seq', inplace=True)
        grouped_skills['exer_id:token'] = grouped_skills['exer_id:token'].astype(str)
        grouped_skills['cpt_seq:token_seq'] = grouped_skills['cpt_seq:token_seq'].astype(str)
        grouped_skills  = grouped_skills.groupby('exer_id:token')['cpt_seq:token_seq'].agg(','.join).reset_index()

        # 合并 assignment_id
        grouped_assignments = df_exer[['exer_id:token','assignment_id:token_seq']]
        grouped_assignments.drop_duplicates(inplace=True)
        grouped_assignments.sort_values(by='assignment_id:token_seq', inplace=True)
        grouped_assignments['exer_id:token'] = grouped_assignments['exer_id:token'].astype(str)
        grouped_assignments['assignment_id:token_seq'] = grouped_assignments['assignment_id:token_seq'].astype(str)
        grouped_assignments  = grouped_assignments.groupby('exer_id:token')['assignment_id:token_seq'].agg(','.join).reset_index()

        # 合并结果
        df_exer = pd.merge(grouped_skills, grouped_assignments, on='exer_id:token', how='left')
        df_exer['exer_id:token'] = df_exer['exer_id:token'].astype(int)
        df_exer.sort_values(by='exer_id:token', inplace=True)
        

        # # Save MidData
        # 
        # 此处将数据保存到`self.midpath`中

        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        df_user.to_csv(f"{self.midpath}/{self.dt}.user.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')
        # pd.set_option("mode.chained_assignment", "warn")  # igore warning


from .base_mid2cache import BaseMid2Cache
from .gen_q_mat import M2C_GenQMat
from .label2int import M2C_Label2Int
from .merge_divided_splits import M2C_MergeDividedSplits
from .remapid import M2C_ReMapId
from .build_cpt_relation import M2C_BuildKCRelation
from .build_missing_Q import M2C_BuildMissingQ
from .fill_missing_Q import M2C_FillMissingQ
from .filtering_records_by_attr import M2C_FilteringRecordsByAttr



import networkx as nx
import numpy as np

from .KTInterExtendsQDataTPL import KTInterExtendsQDataTPL
import torch


class QDKTDataTPL(KTInterExtendsQDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_ReMapId', 'M2C_BuildSeqInterFeats', 'M2C_RandomDataSplit4KT', 'M2C_GenKCSeq','M2C_GenQMat','M2C_QDKT_OP'],
    }

    def get_extra_data(self, **kwargs):
        return {
            'laplacian_matrix': self.final_kwargs['laplacian_matrix'],
            'train_dict': self.train_dict
        }

    def set_info_for_fold(self, fold_id):
        super().set_info_for_fold(fold_id)
        self.train_dict = self.dict_train_folds[fold_id]



from edustudio.quickstart import run_edustudio

class TestRun:
    def test_cd(self):
        print("---------- test cd ------------")
        run_edustudio(
            dataset='FrcSub',
            cfg_file_name=None,
            traintpl_cfg_dict={
                'cls': 'GeneralTrainTPL',
                'epoch_num': 2,
                'device': 'cpu'
            },
            datatpl_cfg_dict={
                'cls': 'CDInterExtendsQDataTPL'
            },
            modeltpl_cfg_dict={
                'cls': 'KaNCD',
            },
            evaltpl_cfg_dict={
                'clses': ['PredictionEvalTPL', 'InterpretabilityEvalTPL'],
            }
        )



from .gd_traintpl import GDTrainTPL
from edustudio.utils.common import UnifyConfig, set_same_seeds, tensor2npy
from edustudio.utils.callback import ModelCheckPoint, EarlyStopping, History, BaseLogger, Callback, CallbackList
from edustudio.model import BaseModel
import torch
from torch.optim import lr_scheduler
from torch.autograd import Variable, grad
from typing import Sequence
from collections import defaultdict
from tqdm import tqdm
import numpy as np
import shutil


class KTLoss(torch.nn.Module):

    def __init__(self):
        super(KTLoss, self).__init__()

    def forward(self, pred_answers, real_answers, mask_seq):

        real_answers = real_answers[:, 1:]
        answer_mask = mask_seq.long()
        
        y_pred = pred_answers[answer_mask].float()
        y_true = real_answers[answer_mask].float()
        
        loss=torch.nn.BCELoss()(y_pred, y_true)
        return loss, y_pred, y_true

def _l2_normalize_adv(d):
    if isinstance(d, Variable):
        d = d.data.cpu().numpy()
    elif isinstance(d, torch.FloatTensor) or isinstance(d, torch.cuda.FloatTensor):
        d = d.cpu().numpy()
    d /= (np.sqrt(np.sum(d ** 2, axis=(1, 2))).reshape((-1, 1, 1)) + 1e-16)
    return torch.from_numpy(d)
    
class AtktTrainTPL(GDTrainTPL):
    default_cfg = {
        'batch_size': 64,
        'eval_batch_size': 64,
        'num_stop_rounds': 10,
        'early_stop_metrics': [('auc','max')],
        'best_epoch_metric': 'auc',
        'unsave_best_epoch_pth': True,
        'ignore_metrics_in_train': [],
        'lr_decay': 50,
        'gamma': 0.5,
        'epsilon': 10,
        'beta': 0.2
    }

    def __init__(self, cfg: UnifyConfig):
        super().__init__(cfg)
        
    def _check_params(self):
        super()._check_params()
        assert self.traintpl_cfg['best_epoch_metric'] in set(i[0] for i in self.traintpl_cfg['early_stop_metrics'])

    def one_fold_start(self, fold_id):
        super().one_fold_start(fold_id)
        # callbacks
        num_stop_rounds = self.traintpl_cfg['num_stop_rounds']
        es_metrics = self.traintpl_cfg["early_stop_metrics"]
        modelCheckPoint = ModelCheckPoint(
            es_metrics, save_folder_path=f"{self.frame_cfg.temp_folder_path}/pths/{fold_id}/"
        )
        es_cb = EarlyStopping(es_metrics, num_stop_rounds=num_stop_rounds, start_round=1)
        history_cb = History(folder_path=f"{self.frame_cfg.temp_folder_path}/history/{fold_id}", plot_curve=True)
        callbacks = [
            modelCheckPoint, es_cb, history_cb,
            BaseLogger(self.logger, group_by_contains=['loss'])
        ]
        self.callback_list = CallbackList(callbacks=callbacks, model=self.model, logger=self.logger)
        # evaltpls
        for evaltpl in self.evaltpls: 
            evaltpl.set_callback_list(self.callback_list)
            evaltpl.set_dataloaders(train_loader=self.train_loader, 
                                    valid_loader=self.valid_loader, 
                                    test_loader=self.test_loader
                                    )
        # train
        set_same_seeds(self.traintpl_cfg['seed'])
        if self.valid_loader is not None:
            self.fit(train_loader=self.train_loader, valid_loader=self.valid_loader)
        else:
            self.fit(train_loader=self.train_loader, valid_loader=self.test_loader)
        
        metric_name = self.traintpl_cfg['best_epoch_metric']
        metric = [m for m in modelCheckPoint.metric_list if m.name == metric_name][0]
        if self.valid_loader is not None:
            # load best params
            fpth =  f"{self.frame_cfg.temp_folder_path}/pths/{fold_id}/best-epoch-{metric.best_epoch:03d}-for-{metric.name}.pth"
            self.model.load_state_dict(torch.load(fpth))

            metrics = self.inference(self.test_loader)
            for name in metrics: self.logger.info(f"{name}: {metrics[name]}")
            History.dump_json(metrics, f"{self.frame_cfg.temp_folder_path}/{fold_id}/result.json")
        else:
            metrics = history_cb.log_as_time[metric.best_epoch]

        if self.traintpl_cfg['unsave_best_epoch_pth']: shutil.rmtree(f"{self.frame_cfg.temp_folder_path}/pths/")
        return metrics

    def fit(self, train_loader, valid_loader):
        kt_loss = KTLoss()
        self.model.train()
        optimizer = self.traintpl_cfg['optim']
        lr = self.traintpl_cfg['lr']
        weight_decay = self.traintpl_cfg['weight_decay']
        eps = self.traintpl_cfg['eps']
        self.optimizer = self._get_optim(optimizer=optimizer, lr=lr, weight_decay=weight_decay, eps=eps)

        self.scheduler = lr_scheduler.StepLR(self.optimizer, step_size=self.traintpl_cfg['lr_decay'], gamma=self.traintpl_cfg['gamma'])
        self.callback_list.on_train_begin()
        for epoch in range(self.traintpl_cfg['epoch_num']):
            self.callback_list.on_epoch_begin(epoch + 1)
            logs = defaultdict(lambda: np.full((len(train_loader),), np.nan, dtype=np.float32))
            for batch_id, batch_dict in enumerate(
                    tqdm(train_loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[EPOCH={:03d}]".format(epoch + 1))
            ):
                batch_dict = self.batch_dict2device(batch_dict)
                pred_res, features = self.model(**batch_dict)
                loss, y_pred, y_true = kt_loss(pred_res, batch_dict['label_seq'], batch_dict['mask_seq'])

                features_grad = grad(loss, features, retain_graph=True)  # 返回loss对features的梯度
                p_adv = torch.FloatTensor(self.traintpl_cfg['epsilon'] * _l2_normalize_adv(features_grad[0].data))
                p_adv = Variable(p_adv)
                pred_res, features = self.model(**batch_dict, p_adv = p_adv.to(self.traintpl_cfg['device']))
                adv_loss, _ , _ = kt_loss(pred_res, batch_dict['label_seq'], batch_dict['mask_seq'])

                total_loss = loss + self.traintpl_cfg['beta']*adv_loss
                self.optimizer.zero_grad()
                total_loss.backward()
                self.optimizer.step()

                loss_dict = {'total_loss': total_loss, 'loss': loss, 'adv_loss': adv_loss}
                for k in loss_dict: logs[k][batch_id] = loss_dict[k].item() if loss_dict[k] is not None else np.nan
            
            self.scheduler.step()

            for name in logs: logs[name] = float(np.nanmean(logs[name]))

            if valid_loader is not None:
                val_metrics = self.evaluate(valid_loader)
                logs.update({f"{metric}": val_metrics[metric] for metric in val_metrics})

            self.callback_list.on_epoch_end(epoch + 1, logs=logs)
            if self.model.share_callback_dict.get('stop_training', False):
                break
            
        self.callback_list.on_train_end()
    
    @torch.no_grad()
    def evaluate(self, loader):
        self.model.eval()
        pd_list = list(range(len(loader)))
        gt_list = list(range(len(loader)))
        for idx, batch_dict in enumerate(tqdm(loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[PREDICT]")):
            batch_dict = self.batch_dict2device(batch_dict)
            eval_dict = self.model.predict(**batch_dict)
            pd_list[idx] = eval_dict['y_pd']
            gt_list[idx] = eval_dict['y_gt']
        y_pd = torch.hstack(pd_list)
        y_gt = torch.hstack(gt_list)

        eval_data_dict = {
            'y_pd': y_pd,
            'y_gt': y_gt,
        }
        if hasattr(self.model, 'get_stu_status'):
            stu_stats_list = []
            idx = torch.arange(0, self.datatpl_cfg['dt_info']['stu_count']).to(self.traintpl_cfg['device'])
            for i in range(0,self.datatpl_cfg['dt_info']['stu_count'], self.traintpl_cfg['eval_batch_size']):
                batch_stu_id = idx[i:i+self.traintpl_cfg['eval_batch_size']]
                batch = self.model.get_stu_status(batch_stu_id)
                stu_stats_list.append(batch)
            stu_stats = torch.vstack(stu_stats_list)
            eval_data_dict.update({
                'stu_stats': tensor2npy(stu_stats),
            })
        if hasattr(loader.dataset, 'Q_mat'):
            eval_data_dict.update({
                'Q_mat': tensor2npy(loader.dataset.Q_mat)
            })
        eval_result = {}
        for evaltpl in self.evaltpls: eval_result.update(
                evaltpl.eval(ignore_metrics=self.traintpl_cfg['ignore_metrics_in_train'], **eval_data_dict)
            )
        return eval_result

    @torch.no_grad()
    def inference(self, loader):
        self.model.eval()
        pd_list = list(range(len(loader)))
        gt_list = list(range(len(loader)))
        for idx, batch_dict in enumerate(tqdm(loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[PREDICT]")):
            batch_dict = self.batch_dict2device(batch_dict)
            eval_dict = self.model.predict(**batch_dict)
            pd_list[idx] = eval_dict['y_pd']
            gt_list[idx] = eval_dict['y_gt']
        y_pd = torch.hstack(pd_list)
        y_gt = torch.hstack(gt_list)

        eval_data_dict = {
            'y_pd': y_pd,
            'y_gt': y_gt,
        }
        if hasattr(self.model, 'get_stu_status'):
            stu_stats_list = []
            idx = torch.arange(0, self.datatpl_cfg['dt_info']['stu_count']).to(self.traintpl_cfg['device'])
            for i in range(0,self.datatpl_cfg['dt_info']['stu_count'], self.traintpl_cfg['eval_batch_size']):
                batch_stu_id = idx[i:i+self.traintpl_cfg['eval_batch_size']]
                batch = self.model.get_stu_status(batch_stu_id)
                stu_stats_list.append(batch)
            stu_stats = torch.vstack(stu_stats_list)
            eval_data_dict.update({
                'stu_stats': tensor2npy(stu_stats),
            })
        if hasattr(loader.dataset, 'Q_mat'):
            eval_data_dict.update({
                'Q_mat': tensor2npy(loader.dataset.Q_mat)
            })
        eval_result = {}
        for evaltpl in self.evaltpls: eval_result.update(evaltpl.eval(**eval_data_dict))
        return eval_result



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from .callbacks.callback import Callback
from .callbacks.earlyStopping import EarlyStopping
from .callbacks.baseLogger import BaseLogger
from .callbacks.modelCheckPoint import ModelCheckPoint
from .callbacks.history import History
from .callbacks.epochPredict import EpochPredict
from .callBackList import CallbackList
from .modeState import ModeState
from .callbacks.tensorboardCallBack import TensorboardCallback



import os
from ..common.edu_datatpl import EduDataTPL
import json
from edustudio.datatpl.common.general_datatpl import DataTPLStatus
import torch


class DCDDataTPL(EduDataTPL):
    default_cfg = {
        'n_folds': 5,
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_FilterRecords4CD', 'M2C_ReMapId', 'M2C_RandomDataSplit4CD', 'M2C_GenQMat', 'M2C_BuildMissingQ', 'M2C_FillMissingQ'],
        'cpt_relation_file_name': 'cpt_relation',
    }

    def __init__(self, cfg, df, df_train=None, df_valid=None, df_test=None, dict_cpt_relation=None, status=DataTPLStatus(), df_stu=None, df_exer=None):
        self.dict_cpt_relation = dict_cpt_relation
        super().__init__(cfg, df, df_train, df_valid, df_test, df_stu, df_exer, status)

    def _check_param(self):
        super()._check_params()
        assert 0 <= self.datatpl_cfg['Q_delete_ratio'] < 1

    @property
    def common_str2df(self):
        dic = super().common_str2df
        dic['dict_cpt_relation'] = self.dict_cpt_relation
        return dic
    

    def process_data(self):
        super().process_data()
        dt_info = self.final_kwargs['dt_info']
        user_count = dt_info['stu_count']
        item_count = dt_info['exer_count']
        self.interact_mat_list = []
        for interact_df in self.final_kwargs['df_train_folds']:
            interact_mat = torch.zeros((user_count, item_count), dtype=torch.int8)
            idx = interact_df[interact_df['label:float'] == 1][['stu_id:token','exer_id:token']].to_numpy()
            interact_mat[idx[:,0], idx[:,1]] = 1  
            idx = interact_df[interact_df['label:float'] != 1][['stu_id:token','exer_id:token']].to_numpy()
            interact_mat[idx[:,0], idx[:,1]] = -1
            self.interact_mat_list.append(interact_mat)

        self.final_kwargs['interact_mat_list'] = self.interact_mat_list

        if self.final_kwargs['dict_cpt_relation'] is None:
            self.final_kwargs['dict_cpt_relation'] = {i: [i] for i in range(self.final_kwargs['dt_info']['cpt_count'])}

    @classmethod
    def load_data(cls, cfg):
        kwargs = super().load_data(cfg)
        fph = f"{cfg.frame_cfg.data_folder_path}/middata/{cfg.datatpl_cfg['cpt_relation_file_name']}.json"
        if os.path.exists(fph):
            with open(fph, 'r', encoding='utf-8') as f:
                kwargs['dict_cpt_relation'] = json.load(f)
        else:
            cfg.logger.warning("without cpt_relation.json")
            kwargs['dict_cpt_relation'] = None
        return kwargs

    def get_extra_data(self):
        extra_dict = super().get_extra_data()
        extra_dict['filling_Q_mat'] = self.filling_Q_mat
        extra_dict['interact_mat'] = self.interact_mat
        return extra_dict

    def set_info_for_fold(self, fold_id):
        super().set_info_for_fold(fold_id)
        self.filling_Q_mat = self.final_kwargs['filling_Q_mat_list'][fold_id]
        self.interact_mat = self.final_kwargs['interact_mat_list'][fold_id]



from .build_seq_inter_feats import M2C_BuildSeqInterFeats
from .cpt_as_exer import M2C_KCAsExer
from .gen_cpt_seq import M2C_GenKCSeq
from .gen_unfold_cpt_seq import M2C_GenUnFoldKCSeq
from .data_split4kt import M2C_RandomDataSplit4KT



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from .callback import Callback
import os
import pickle


class EpochPredict(Callback):
    def __init__(self, save_folder_path, fmt="predict-{epoch}", **kwargs):
        super(EpochPredict, self).__init__()
        self.kwargs = kwargs
        self.save_folder_path = save_folder_path
        if not os.path.exists(save_folder_path):
            os.makedirs(self.save_folder_path)
        self.file_fmt = fmt + ".pkl"
        self.pd = None
        assert "{epoch}" in self.file_fmt

    def on_epoch_end(self, epoch, logs=None, **kwargs):
        self.pd = self.model.predict(**self.kwargs)
        filepath = os.path.join(self.save_folder_path, self.file_fmt.format(epoch=epoch))
        self.to_pickle(filepath, self.pd)

    @staticmethod
    def to_pickle(filepath, obj):
        with open(filepath, 'wb') as f:
            pickle.dump(obj, f, protocol=4)



from ..common import BaseMid2Cache
import pandas as pd
import numpy as np


class M2C_MGCD_OP(BaseMid2Cache):
    default_cfg = {
        "group_id_field": "class_id:token",
        "min_inter": 5,
    }

    def __init__(self, m2c_cfg, is_dataset_divided) -> None:
        super().__init__(m2c_cfg)
        self.is_dataset_divided = is_dataset_divided
        assert self.is_dataset_divided is False

    @classmethod
    def from_cfg(cls, cfg):
        m2c_cfg = cfg.datatpl_cfg.get(cls.__name__)
        is_dataset_divided = cfg.datatpl_cfg.is_dataset_divided
        return cls(m2c_cfg, is_dataset_divided)

    def process(self, **kwargs):
        df_stu = kwargs['df_stu']
        df_inter = kwargs['df']
        df_inter_group, df_inter_stu, kwargs['df_stu']= self.get_df_group_and_df_inter(
            df_inter=df_inter, df_stu=df_stu
        )
        kwargs['df'] = df_inter_group
        kwargs['df_inter_stu'] = df_inter_stu
        kwargs['df_stu'] = kwargs['df_stu'].rename(columns={self.m2c_cfg['group_id_field']: 'group_id:token'})
        self.group_count = df_inter_group['group_id:token'].nunique()
        return kwargs

    def get_df_group_and_df_inter(self, df_stu:pd.DataFrame, df_inter:pd.DataFrame):
        df_inter = df_inter.merge(df_stu[['stu_id:token', self.m2c_cfg['group_id_field']]], on=['stu_id:token'], how='left')  # 两个表合并，这样inter里也有class_id

        df_inter_group = pd.DataFrame()
        df_inter_stu = pd.DataFrame()

        for _, inter_g in df_inter.groupby(self.m2c_cfg['group_id_field']):
            exers_list = inter_g[['stu_id:token', 'exer_id:token']].groupby('stu_id:token').agg(set)['exer_id:token'].tolist()
            inter_exer_set = None # 选择所有学生都做的题目
            for exers in exers_list: inter_exer_set = exers if inter_exer_set is None else (inter_exer_set & exers)
            inter_exer_set = np.array(list(inter_exer_set))
            if inter_exer_set.shape[0] >= self.m2c_cfg['min_inter']:
                tmp_group_df = inter_g[inter_g['exer_id:token'].isin(inter_exer_set)]
                tmp_stu_df = inter_g[~inter_g['exer_id:token'].isin(inter_exer_set)]

                df_inter_group = pd.concat([df_inter_group, tmp_group_df], ignore_index=True, axis=0)
                df_inter_stu = pd.concat([df_inter_stu, tmp_stu_df], ignore_index=True, axis=0)
            else:
                df_stu = df_stu[df_stu['class_id:token'] != inter_g['class_id:token'].values[0]]

        df_inter_group = df_inter_group[['label:float', 'exer_id:token', self.m2c_cfg['group_id_field']]].groupby(
            [self.m2c_cfg['group_id_field'], 'exer_id:token']
        ).agg('mean').reset_index().rename(columns={self.m2c_cfg['group_id_field']: 'group_id:token'})

        return df_inter_group, df_inter_stu, df_stu[df_stu['class_id:token'].isin(df_inter_group['group_id:token'].unique())]
    
    def set_dt_info(self, dt_info, **kwargs):
        dt_info['group_count'] = self.group_count


from ..KT import M2C_BuildSeqInterFeats
import torch 
import pandas as pd
import numpy as np
from edustudio.datatpl.utils import PadSeqUtil
from collections import defaultdict


class M2C_CL4KT_OP(M2C_BuildSeqInterFeats):
    default_cfg = {
        'seed': 2023,
        'divide_by': 'stu',
        'window_size': 200,
        "divide_scale_list": [7,1,2],
        'sequence_truncation': '', # option: recent、history

    }
    def __init__(self, m2c_cfg, n_folds, is_dataset_divided) -> None:
        super().__init__(m2c_cfg, n_folds, is_dataset_divided)
    
    def process(self, **kwargs):
        kwargs = super().process(**kwargs)
        kwargs = self.compute_cpt2difflevel(**kwargs)

        return kwargs

    def compute_cpt2difflevel(self, **kwargs):
        cpt_correct = defaultdict(int)
        cpt_count = defaultdict(int)
        for i, (c_list, r_list) in enumerate(zip(kwargs['df_seq']['cpt_unfold_seq:token_seq'], kwargs['df_seq']['label_seq:float_seq'])):
            for c, r in zip(c_list[kwargs['df_seq']['mask_seq:token_seq'][i] == 1], r_list[kwargs['df_seq']['mask_seq:token_seq'][i] == 1]):
                cpt_correct[c] += r
                cpt_count[c] += 1
        cpt_diff = {c: cpt_correct[c] / float(cpt_count[c]) for c in cpt_correct}  # cpt difficult
        ordered_cpts = [item[0] for item in sorted(cpt_diff.items(), key=lambda x: x[1])]
        easier_cpts, harder_cpts = defaultdict(int), defaultdict(int)
        for index, cpt in enumerate(ordered_cpts):  
            if index == 0:
                easier_cpts[cpt] = ordered_cpts[index + 1]
                harder_cpts[cpt] = cpt
            elif index == len(ordered_cpts) - 1:
                easier_cpts[cpt] = cpt
                harder_cpts[cpt] = ordered_cpts[index - 1]
            else:
                easier_cpts[cpt] = ordered_cpts[index + 1]
                harder_cpts[cpt] = ordered_cpts[index - 1]

        kwargs['easier_cpts'] = easier_cpts
        kwargs['harder_cpts'] = harder_cpts

        return kwargs

    def construct_df2dict(self, df: pd.DataFrame):
        if df is None: return None
     
        if self.m2c_cfg['sequence_truncation'] == 'recent':
            tmp_df = df[['stu_id:token', 'exer_id:token', 'cpt_unfold:token', 'label:float']].groupby('stu_id:token').agg(
                lambda x: list(x)[-self.m2c_cfg['window_size']:]
            ).reset_index()

            exer_seq, idx, mask_seq = PadSeqUtil.pad_sequence(  # mask_seq corresponding to attention_mask
                tmp_df['exer_id:token'].to_list(), return_idx=True, return_mask=True,
                maxlen=self.m2c_cfg['window_size'], padding='pre'
            )
            cpt_unfold_seq, _, _ = PadSeqUtil.pad_sequence(
                tmp_df['cpt_unfold:token'].to_list(),
                maxlen=self.m2c_cfg['window_size'], padding='pre'
            )
            label_seq, _, _ = PadSeqUtil.pad_sequence(
                tmp_df['label:float'].to_list(), dtype=np.float32,
                maxlen=self.m2c_cfg['window_size'], padding='pre'
            )
            
        else:
            tmp_df = df[['stu_id:token', 'exer_id:token', 'cpt_unfold:token', 'label:float']].groupby('stu_id:token').agg(
                lambda x: list(x)[:self.m2c_cfg['window_size']]
            ).reset_index()
            exer_seq, idx, mask_seq = PadSeqUtil.pad_sequence(  # mask_seq corresponding to attention_mask
                tmp_df['exer_id:token'].to_list(), return_idx=True, return_mask=True,
                maxlen=self.m2c_cfg['window_size'],
            )
            cpt_unfold_seq, _, _ = PadSeqUtil.pad_sequence(
                tmp_df['cpt_unfold:token'].to_list(),
                maxlen=self.m2c_cfg['window_size'],
            )
            label_seq, _, _ = PadSeqUtil.pad_sequence(
                tmp_df['label:float'].to_list(), dtype=np.float32,
                maxlen=self.m2c_cfg['window_size'],
            )


        stu_id = tmp_df['stu_id:token'].to_numpy()[idx]

        ret_dict = {
            'stu_id:token': stu_id,
            'exer_seq:token_seq': exer_seq,
            'cpt_unfold_seq:token_seq': cpt_unfold_seq,
            'label_seq:float_seq': label_seq,
            'mask_seq:token_seq': mask_seq
        }

        return ret_dict
    
    def set_dt_info(self, dt_info, **kwargs):
        super().set_dt_info(dt_info, **kwargs)
        dt_info['train_easier_cpts'] = kwargs['easier_cpts']
        dt_info['train_harder_cpts'] = kwargs['easier_cpts']



import pandas as pd
import os
import time
from .raw2mid import BaseRaw2Mid
from sklearn.preprocessing import LabelEncoder

r"""
R2M_JunyiExerAsCpt
########################
"""


class R2M_JunyiExerAsCpt(BaseRaw2Mid):
    """R2M_JunyiExerAsCpt is a class used to handle the Junyi dataset, where we consider the exercise's name as the basis for constructing the cpt_seq (concept sequence)."""

    def process(self, **kwargs):
        super().process()
        # # 处理 Junyi Problem log dataset
        data = pd.read_csv(f"{self.rawpath}/junyi_ProblemLog_original.csv", encoding='utf-8', low_memory=True)
        # # 去除花费时间小于等于0s的或者超过半年的6*30*24*60*60s = 15552000s
        # data.drop(data[(data.time_taken <= 0) | (data.time_taken >= 15552000)].index, inplace=True)
        data = data[['user_id', 'exercise', 'time_done', 'time_taken', 'correct']]
        # 确定start_timestamp:float, order_id:float, correct 和 time_taken (单位时间为ms)
        data['start_timestamp:float'] = data['time_done'] // 1000 - data['time_taken'] * 1000
        data = data.sort_values(by='start_timestamp:float', ascending=True)
        data['order_id:float'] = range(len(data))
        data['correct'] = data['correct'].astype(int)
        data['time_taken'] = data['time_taken'] * 1000
        df_inter = data.rename(columns={
            'user_id': 'stu_id:token', 'exercise': 'exer_id:token', 'time_taken': 'cost_time:float',
            'correct': 'label:float', 'time_done': 'time_done:float'
        })

        # read exer info
        raw_df_exer = pd.read_csv(
            f"{self.rawpath}/junyi_Exercise_table.csv", encoding='utf-8', low_memory=False,
            usecols=['name', 'prerequisites']
        )

        df_relation = raw_df_exer.copy()
        df_relation.dropna(inplace=True)
        df_relation.rename(columns={"name": "cpt_tail:token", 'prerequisites': "cpt_head:token"}, inplace=True)
        df_relation['cpt_head:token'] = df_relation['cpt_head:token'].apply(lambda x: x.split(","))
        df_relation = df_relation[['cpt_head:token', 'cpt_tail:token']].explode('cpt_head:token')

        df_exer = raw_df_exer.rename(columns={'name': 'exer_id:token', }) # 'topic': 'topic:token', 'area': 'area:token'

        exer_lbe = LabelEncoder()
        exer_lbe.fit(df_exer['exer_id:token'].tolist() + df_inter['exer_id:token'].tolist())
        df_inter['exer_id:token'] = exer_lbe.transform(df_inter['exer_id:token'])
        df_exer['exer_id:token'] = exer_lbe.transform(df_exer['exer_id:token'])
        df_relation["cpt_tail:token"] = exer_lbe.transform(df_relation["cpt_tail:token"])
        df_relation["cpt_head:token"] = exer_lbe.transform(df_relation["cpt_head:token"])

        df_exer.drop(columns=['prerequisites'], inplace=True)
        df_exer['cpt_seq:token_seq'] = df_exer['exer_id:token'].astype(str)
        df_exer.drop_duplicates('exer_id:token', inplace=True)
        
        # 此处将数据保存到`self.midpath`中
        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        # df_stu.to_csv(f"{self.midpath}/{self.dt}.stu.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')
        df_relation.to_csv(f"{self.midpath}/{self.dt}.cpt_relation.prerequisite.csv", index=False, encoding='utf-8')



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os

from setuptools import setup, find_packages

install_requires = [
    "torch>=1.10.0",
    "numpy>=1.17.2",
    "scipy>=1.6.0",
    "pandas>=1.0.5",
    "tqdm>=4.48.2",
    "scikit_learn>=0.23.2",
    "pyyaml>=5.1.0",
    "tensorboard>=2.5.0",
    "requests>=2.27.1",
    "pytz>=2022.1",
    "matplotlib>=3.5.1",
    "deepdiff>=6.3.1",
    "networkx>=2.8"
]

setup_requires = []

extras_require = {}

classifiers = ["License :: OSI Approved :: MIT License"]

long_description = (
    "EduStudio is a Unified and Templatized Framework "
    "for Student Assessment Models including "
    "Cognitive Diagnosis(CD) and Knowledge Tracing(KT) based on Pytorch."
)

# Readthedocs requires Sphinx extensions to be specified as part of
# install_requires in order to build properly.
on_rtd = os.environ.get("READTHEDOCS", None) == "True"
if on_rtd:
    install_requires.extend(setup_requires)

setup(
    name="edustudio",
    version="v1.1.6", 
    description="a Unified and Templatized Framework for Student Assessment Models",
    long_description=long_description,
    python_requires='>=3.8',
    long_description_content_type="text/markdown",
    url="https://github.com/HFUT-LEC/EduStudio",
    author="HFUT-LEC",
    author_email="lmcRS.hfut@gmail.com",
    packages=[package for package in find_packages() if package.startswith("edustudio")],
    include_package_data=True,
    install_requires=install_requires,
    setup_requires=setup_requires,
    extras_require=extras_require,
    zip_safe=False,
    classifiers=classifiers,
    entry_points={
        "console_scripts": [
            "edustudio = edustudio.quickstart.atom_cmds:entrypoint",
        ],
    }
)



from ..common.base_mid2cache import BaseMid2Cache
import pandas as pd


class M2C_FilterRecords4CD(BaseMid2Cache):
    default_cfg = {
        "stu_least_records": 10,
        "exer_least_records": 0,
    }

    def process(self, **kwargs):
        df: pd.DataFrame = kwargs['df']
        assert df is not None

        # 去重，保留第一个记录
        df.drop_duplicates(
            subset=['stu_id:token', 'exer_id:token', "label:float"], 
            keep='first', inplace=True, ignore_index=True
        )
        
        stu_least_records = self.m2c_cfg['stu_least_records']
        exer_least_records = self.m2c_cfg['exer_least_records']

        # 循环删除user和item
        last_count = 0
        while last_count != df.__len__():
            last_count = df.__len__()
            gp_by_uid = df[['stu_id:token','exer_id:token']].groupby('stu_id:token').agg('count').reset_index()
            selected_users = gp_by_uid[gp_by_uid['exer_id:token'] >= stu_least_records].reset_index()['stu_id:token'].to_numpy()

            gp_by_iid = df[['stu_id:token','exer_id:token']].groupby('exer_id:token').agg('count').reset_index()
            selected_items = gp_by_iid[gp_by_iid['stu_id:token'] >= exer_least_records].reset_index()['exer_id:token'].to_numpy()

            df = df[df['stu_id:token'].isin(selected_users) & df['exer_id:token'].isin(selected_items)]


        df = df.reset_index(drop=True)
        selected_users = df['stu_id:token'].unique()
        selected_items = df['exer_id:token'].unique()

        if kwargs.get('df_exer', None) is not None:
            kwargs['df_exer'] = kwargs['df_exer'][kwargs['df_exer']['exer_id:token'].isin(selected_items)]
            kwargs['df_exer'].reset_index(drop=True, inplace=True)

        if kwargs.get('df_stu', None) is not None:
            kwargs['df_stu'] = kwargs['df_stu'][kwargs['df_stu']['stu_id:token'].isin(selected_users)]
            kwargs['df_stu'].reset_index(drop=True, inplace=True)

        kwargs['df'] = df
        return kwargs



from ..common import EduDataTPL

class KTInterCptAsExerDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ["M2C_KCAsExer", 'M2C_Label2Int', 'M2C_ReMapId', 'M2C_BuildSeqInterFeats', 'M2C_RandomDataSplit4KT'],
    }

